"""Callback-based anomaly handler."""

from __future__ import annotations

import logging
from collections.abc import Callable
from typing import TYPE_CHECKING

import pandas as pd

from finlab_sentinel.handlers.base import AnomalyHandler

if TYPE_CHECKING:
    from finlab_sentinel.comparison.report import AnomalyReport

logger = logging.getLogger(__name__)


class CallbackHandler(AnomalyHandler):
    """Handler that executes callback function on anomaly.

    The callback is called with the AnomalyReport, then the fallback
    handler determines what data to return.
    """

    def __init__(
        self,
        callback: Callable[[AnomalyReport], None],
        fallback_handler: AnomalyHandler,
    ) -> None:
        """Initialize callback handler.

        Args:
            callback: Function to call with anomaly report
            fallback_handler: Handler to use for data return decision
        """
        self.callback = callback
        self.fallback_handler = fallback_handler

    @property
    def name(self) -> str:
        return f"callback+{self.fallback_handler.name}"

    def handle(
        self,
        report: AnomalyReport,
        cached_data: pd.DataFrame,
        new_data: pd.DataFrame,
    ) -> pd.DataFrame:
        """Execute callback and delegate to fallback handler.

        Args:
            report: Anomaly report
            cached_data: Previously cached DataFrame
            new_data: New DataFrame

        Returns:
            DataFrame from fallback handler
        """
        # Execute callback
        try:
            self.callback(report)
        except Exception as e:
            # Log but don't fail on callback errors
            logger.error(
                f"Callback failed for anomaly in {report.dataset}: {e}",
                exc_info=True,
            )

        # Delegate to fallback handler
        return self.fallback_handler.handle(report, cached_data, new_data)


def create_handler_from_config(
    behavior: str,
    callback: Callable[[AnomalyReport], None] | None = None,
) -> AnomalyHandler:
    """Create handler from configuration.

    Args:
        behavior: Behavior string from config
        callback: Optional callback function

    Returns:
        Configured handler
    """
    from finlab_sentinel.handlers.exception import RaiseExceptionHandler
    from finlab_sentinel.handlers.warning import (
        WarnReturnCachedHandler,
        WarnReturnNewHandler,
    )

    # Create base handler
    base_handler: AnomalyHandler
    if behavior == "raise":
        base_handler = RaiseExceptionHandler()
    elif behavior == "warn_return_cached":
        base_handler = WarnReturnCachedHandler()
    elif behavior == "warn_return_new":
        base_handler = WarnReturnNewHandler()
    else:
        # Default to raise
        logger.warning(f"Unknown behavior '{behavior}', defaulting to 'raise'")
        base_handler = RaiseExceptionHandler()

    # Wrap with callback if provided
    if callback is not None:
        return CallbackHandler(callback, base_handler)

    return base_handler



"""Tests for cleanup utilities."""

from datetime import datetime, timedelta
from pathlib import Path

import pandas as pd
import pytest

from finlab_sentinel.config.schema import SentinelConfig, StorageConfig
from finlab_sentinel.storage.cleanup import (
    calculate_directory_size,
    cleanup_on_startup,
    get_storage_info,
)
from finlab_sentinel.storage.parquet import ParquetStorage


@pytest.fixture
def config_with_storage(tmp_path: Path) -> SentinelConfig:
    """Create config with temp storage path."""
    return SentinelConfig(storage=StorageConfig(path=tmp_path, retention_days=7))


@pytest.fixture
def config_with_min_backups(tmp_path: Path) -> SentinelConfig:
    """Create config with custom min_backups_per_dataset."""
    return SentinelConfig(
        storage=StorageConfig(
            path=tmp_path, retention_days=7, min_backups_per_dataset=2
        )
    )


class TestCleanupOnStartup:
    """Tests for cleanup_on_startup function."""

    def test_no_expired_backups(
        self, config_with_storage: SentinelConfig, sample_df: pd.DataFrame
    ):
        """Verify no cleanup when no expired backups."""
        # Create a recent backup
        storage = ParquetStorage(
            base_path=config_with_storage.get_storage_path(),
            compression=config_with_storage.storage.compression,
        )
        storage.save("test", "test:dataset", sample_df, "hash123")

        # Should not delete anything
        deleted = cleanup_on_startup(config_with_storage)
        assert deleted == 0

    def test_returns_zero_for_empty_storage(self, config_with_storage: SentinelConfig):
        """Verify returns 0 when storage is empty."""
        deleted = cleanup_on_startup(config_with_storage)
        assert deleted == 0

    def test_keeps_min_backups_when_all_expired(
        self, config_with_min_backups: SentinelConfig, sample_df: pd.DataFrame
    ):
        """Verify cleanup keeps min_backups_per_dataset even if all expired."""
        storage = ParquetStorage(
            base_path=config_with_min_backups.get_storage_path(),
            compression=config_with_min_backups.storage.compression,
        )

        # Create 4 backups all older than retention_days (spaced by minutes for different filenames)
        base_time = datetime.now() - timedelta(days=30)
        for i in range(4):
            storage.save("test", "test:dataset", sample_df, f"hash{i}")
            # Manually set old timestamp (each 10 minutes apart)
            with storage.index._connect() as conn:
                old_date = (base_time - timedelta(minutes=i * 10)).isoformat()
                conn.execute(
                    "UPDATE backups SET created_at = ? WHERE content_hash = ?",
                    (old_date, f"hash{i}"),
                )

        # Cleanup should keep at least min_backups_per_dataset=2
        cleanup_on_startup(config_with_min_backups)

        # Should have 2 remaining in index
        remaining = storage.list_backups("test")
        assert len(remaining) == 2

    def test_respects_min_backups_config(self, tmp_path: Path, sample_df: pd.DataFrame):
        """Verify cleanup uses min_backups_per_dataset from config."""
        config = SentinelConfig(
            storage=StorageConfig(
                path=tmp_path,
                retention_days=7,
                min_backups_per_dataset=5,  # Keep at least 5
            )
        )
        storage = ParquetStorage(
            base_path=config.get_storage_path(),
            compression=config.storage.compression,
        )

        # Create 7 old backups (spaced by minutes for different filenames)
        base_time = datetime.now() - timedelta(days=30)
        for i in range(7):
            storage.save("test", "test:dataset", sample_df, f"hash{i}")
            with storage.index._connect() as conn:
                old_date = (base_time - timedelta(minutes=i * 10)).isoformat()
                conn.execute(
                    "UPDATE backups SET created_at = ? WHERE content_hash = ?",
                    (old_date, f"hash{i}"),
                )

        # Cleanup should keep at least min_backups_per_dataset=5
        cleanup_on_startup(config)

        # Should have 5 remaining in index
        remaining = storage.list_backups("test")
        assert len(remaining) == 5


class TestGetStorageInfo:
    """Tests for get_storage_info function."""

    def test_empty_storage(self, config_with_storage: SentinelConfig):
        """Verify stats for empty storage."""
        stats = get_storage_info(config_with_storage)

        assert stats["total_backups"] == 0
        assert stats["total_size_bytes"] == 0

    def test_with_backups(
        self, config_with_storage: SentinelConfig, sample_df: pd.DataFrame
    ):
        """Verify stats with existing backups."""
        storage = ParquetStorage(
            base_path=config_with_storage.get_storage_path(),
            compression=config_with_storage.storage.compression,
        )
        storage.save("test1", "test:dataset1", sample_df, "hash1")
        storage.save("test2", "test:dataset2", sample_df, "hash2")

        stats = get_storage_info(config_with_storage)

        assert stats["total_backups"] == 2
        assert stats["total_size_bytes"] > 0
        assert stats["unique_datasets"] == 2


class TestCalculateDirectorySize:
    """Tests for calculate_directory_size function."""

    def test_empty_directory(self, tmp_path: Path):
        """Verify size of empty directory is 0."""
        size = calculate_directory_size(tmp_path)
        assert size == 0

    def test_nonexistent_directory(self, tmp_path: Path):
        """Verify nonexistent directory returns 0."""
        nonexistent = tmp_path / "does_not_exist"
        size = calculate_directory_size(nonexistent)
        assert size == 0

    def test_directory_with_files(self, tmp_path: Path):
        """Verify correct size calculation."""
        # Create some files
        file1 = tmp_path / "file1.txt"
        file1.write_text("hello")  # 5 bytes

        file2 = tmp_path / "subdir" / "file2.txt"
        file2.parent.mkdir()
        file2.write_text("world!")  # 6 bytes

        size = calculate_directory_size(tmp_path)
        assert size == 11



"""Main CLI entry point for finlab-sentinel."""

from __future__ import annotations

from datetime import datetime, timedelta
from pathlib import Path

import typer
from rich.console import Console
from rich.table import Table

app = typer.Typer(
    name="sentinel",
    help="finlab-sentinel: Defensive layer for finlab data.get API",
    add_completion=True,
)
console = Console()

# Global options
_config_path: Path | None = None
_verbose: bool = False


@app.callback()
def main(
    config: Path | None = typer.Option(
        None,
        "--config",
        "-c",
        help="Path to configuration file",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Enable verbose output",
    ),
) -> None:
    """finlab-sentinel CLI - Monitor finlab data changes."""
    global _config_path, _verbose
    _config_path = config
    _verbose = verbose

    if verbose:
        import logging

        logging.basicConfig(level=logging.DEBUG)


@app.command("list")
def list_backups(
    dataset: str | None = typer.Option(
        None,
        "--dataset",
        "-d",
        help="Filter by dataset name",
    ),
    days: int | None = typer.Option(
        None,
        "--days",
        help="Show only last N days",
    ),
    format: str = typer.Option(
        "table",
        "--format",
        "-f",
        help="Output format: table, json, csv",
    ),
) -> None:
    """List all backups."""
    from finlab_sentinel.config.loader import load_config
    from finlab_sentinel.storage.parquet import ParquetStorage, sanitize_backup_key

    config = load_config(_config_path)
    storage = ParquetStorage(
        base_path=config.get_storage_path(),
        compression=config.storage.compression,
    )

    # Get backup key if dataset specified
    backup_key = sanitize_backup_key(dataset) if dataset else None

    backups = storage.list_backups(backup_key)

    # Filter by days if specified
    if days:
        cutoff = datetime.now() - timedelta(days=days)
        backups = [b for b in backups if b.created_at >= cutoff]

    if not backups:
        console.print("[yellow]No backups found[/yellow]")
        return

    if format == "json":
        import json

        console.print(json.dumps([b.to_dict() for b in backups], indent=2))
        return

    if format == "csv":
        console.print("dataset,backup_key,created_at,rows,columns,size_bytes")
        for b in backups:
            console.print(
                f"{b.dataset},{b.backup_key},{b.created_at.isoformat()},"
                f"{b.row_count},{b.column_count},{b.file_size_bytes}"
            )
        return

    # Table format
    table = Table(title="Backups")
    table.add_column("Dataset", style="cyan")
    table.add_column("Created At", style="green")
    table.add_column("Rows", justify="right")
    table.add_column("Columns", justify="right")
    table.add_column("Size", justify="right")

    for b in backups:
        size_str = _format_size(b.file_size_bytes)
        table.add_row(
            b.dataset,
            b.created_at.strftime("%Y-%m-%d %H:%M"),
            str(b.row_count),
            str(b.column_count),
            size_str,
        )

    console.print(table)


@app.command("cleanup")
def cleanup(
    days: int | None = typer.Option(
        None,
        "--days",
        "-d",
        help="Retention days (overrides config)",
    ),
    dry_run: bool = typer.Option(
        False,
        "--dry-run",
        help="Show what would be deleted without deleting",
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Skip confirmation prompt",
    ),
) -> None:
    """Clean up expired backups."""
    from finlab_sentinel.config.loader import load_config
    from finlab_sentinel.storage.parquet import ParquetStorage

    config = load_config(_config_path)
    retention_days = days if days is not None else config.storage.retention_days

    storage = ParquetStorage(
        base_path=config.get_storage_path(),
        compression=config.storage.compression,
    )

    # Preview what would be deleted
    cutoff = datetime.now() - timedelta(days=retention_days)
    backups = storage.list_backups()
    to_delete = [b for b in backups if b.created_at < cutoff]

    if not to_delete:
        console.print(f"[green]No backups older than {retention_days} days[/green]")
        return

    # Show what will be deleted
    total_size = sum(b.file_size_bytes for b in to_delete)
    console.print(
        f"Found [bold]{len(to_delete)}[/bold] backups older than "
        f"{retention_days} days ({_format_size(total_size)})"
    )

    if dry_run:
        console.print("[yellow]Dry run - no files deleted[/yellow]")
        for b in to_delete:
            console.print(f"  Would delete: {b.dataset} ({b.created_at})")
        return

    # Confirm unless forced
    if not force:
        confirm = typer.confirm("Delete these backups?")
        if not confirm:
            console.print("[yellow]Aborted[/yellow]")
            return

    # Delete
    deleted = storage.cleanup_expired(retention_days)
    console.print(f"[green]Deleted {deleted} backups[/green]")


@app.command("export")
def export_backup(
    dataset: str = typer.Argument(..., help="Dataset name to export"),
    output: Path = typer.Option(
        Path("."),
        "--output",
        "-o",
        help="Output directory or file",
    ),
    format: str = typer.Option(
        "parquet",
        "--format",
        "-f",
        help="Export format: parquet, csv",
    ),
    date: str | None = typer.Option(
        None,
        "--date",
        help="Export specific date (YYYY-MM-DD)",
    ),
) -> None:
    """Export backup data."""
    from finlab_sentinel.config.loader import load_config
    from finlab_sentinel.storage.parquet import ParquetStorage, sanitize_backup_key

    config = load_config(_config_path)
    storage = ParquetStorage(
        base_path=config.get_storage_path(),
        compression=config.storage.compression,
    )

    backup_key = sanitize_backup_key(dataset)

    # Load backup
    if date:
        target_date = datetime.strptime(date, "%Y-%m-%d")
        result = storage.load_by_date(backup_key, target_date)
    else:
        result = storage.load_latest(backup_key)

    if result is None:
        console.print(f"[red]No backup found for: {dataset}[/red]")
        raise typer.Exit(1)

    df, metadata = result

    # Determine output path
    if output.is_dir():
        ext = "parquet" if format == "parquet" else "csv"
        output_file = output / f"{backup_key}.{ext}"
    else:
        output_file = output

    # Export
    if format == "csv":
        df.to_csv(output_file)
    else:
        df.to_parquet(output_file)

    console.print(f"[green]Exported to: {output_file}[/green]")


@app.command("accept")
def accept(
    dataset: str = typer.Argument(..., help="Dataset name to accept"),
    reason: str | None = typer.Option(
        None,
        "--reason",
        "-r",
        help="Reason for accepting the new data",
    ),
) -> None:
    """Accept current data as new baseline."""
    from finlab_sentinel.config.loader import load_config
    from finlab_sentinel.core.interceptor import accept_current_data

    config = load_config(_config_path)

    success = accept_current_data(dataset, config, reason)

    if success:
        console.print(f"[green]Accepted new data for: {dataset}[/green]")
        if reason:
            console.print(f"  Reason: {reason}")
    else:
        console.print(f"[red]Failed to accept data for: {dataset}[/red]")
        raise typer.Exit(1)


@app.command("diff")
def diff(
    dataset: str = typer.Argument(..., help="Dataset name to diff"),
    from_date: str | None = typer.Option(
        None,
        "--from",
        help="From date (YYYY-MM-DD)",
    ),
    to_date: str | None = typer.Option(
        None,
        "--to",
        help="To date (YYYY-MM-DD)",
    ),
) -> None:
    """Show differences for a dataset."""
    from finlab_sentinel.comparison.differ import DataFrameComparer
    from finlab_sentinel.config.loader import load_config
    from finlab_sentinel.storage.parquet import ParquetStorage, sanitize_backup_key

    config = load_config(_config_path)
    storage = ParquetStorage(
        base_path=config.get_storage_path(),
        compression=config.storage.compression,
    )

    backup_key = sanitize_backup_key(dataset)

    # Load backups
    if from_date and to_date:
        from_dt = datetime.strptime(from_date, "%Y-%m-%d")
        to_dt = datetime.strptime(to_date, "%Y-%m-%d")

        old_result = storage.load_by_date(backup_key, from_dt)
        new_result = storage.load_by_date(backup_key, to_dt)
    else:
        # Compare latest with current data from finlab
        old_result = storage.load_latest(backup_key)
        new_result = None  # Will fetch from finlab

    if old_result is None:
        console.print(f"[red]No backup found for: {dataset}[/red]")
        raise typer.Exit(1)

    old_df, old_metadata = old_result

    if new_result:
        new_df, _ = new_result
    else:
        # Fetch current from finlab
        try:
            from finlab import data

            new_df = data.get(dataset)
        except ImportError as e:
            console.print("[red]finlab not installed[/red]")
            raise typer.Exit(1) from e

    # Compare
    comparer = DataFrameComparer(
        rtol=config.comparison.rtol,
        atol=config.comparison.atol,
        check_dtype=config.comparison.check_dtype,
        check_na_type=config.comparison.check_na_type,
    )

    result = comparer.compare(old_df, new_df)

    # Display results
    if result.is_identical:
        console.print("[green]No differences found[/green]")
        return

    console.print(f"[bold]Changes: {result.summary()}[/bold]\n")

    if result.added_rows:
        console.print(f"[green]+ {len(result.added_rows)} rows added[/green]")
        for row in list(result.added_rows)[:5]:
            console.print(f"    {row}")
        if len(result.added_rows) > 5:
            console.print(f"    ... and {len(result.added_rows) - 5} more")

    if result.deleted_rows:
        console.print(f"[red]- {len(result.deleted_rows)} rows deleted[/red]")
        for row in list(result.deleted_rows)[:5]:
            console.print(f"    {row}")

    if result.added_columns:
        console.print(f"[green]+ {len(result.added_columns)} columns added[/green]")

    if result.deleted_columns:
        console.print(f"[red]- {len(result.deleted_columns)} columns deleted[/red]")

    if result.modified_cells:
        console.print(f"[yellow]~ {len(result.modified_cells)} cells modified[/yellow]")
        for change in result.modified_cells[:5]:
            console.print(f"    {change}")
        if len(result.modified_cells) > 5:
            console.print(f"    ... and {len(result.modified_cells) - 5} more")

    if result.dtype_changes:
        console.print(f"[yellow]~ {len(result.dtype_changes)} dtype changes[/yellow]")
        for dtype_change in result.dtype_changes:
            console.print(f"    {dtype_change}")


@app.command("config")
def config_cmd(
    action: str = typer.Argument(
        "show",
        help="Action: show, validate, init, paths",
    ),
    output: Path | None = typer.Option(
        None,
        "--output",
        "-o",
        help="Output path for init",
    ),
) -> None:
    """Configuration management."""
    if action == "show":
        from finlab_sentinel.config.loader import load_config

        try:
            config = load_config(_config_path)
            console.print(config.model_dump_json(indent=2))
        except Exception as e:
            console.print(f"[red]Error loading config: {e}[/red]")
            raise typer.Exit(1) from e

    elif action == "validate":
        from finlab_sentinel.config.loader import load_config

        try:
            load_config(_config_path)
            console.print("[green]Configuration is valid[/green]")
        except Exception as e:
            console.print(f"[red]Invalid configuration: {e}[/red]")
            raise typer.Exit(1) from e

    elif action == "init":
        from finlab_sentinel.config.loader import create_default_config_file

        path = output if output else Path("sentinel.toml")
        if path.exists():
            console.print(f"[yellow]File already exists: {path}[/yellow]")
            if not typer.confirm("Overwrite?"):
                raise typer.Exit(0)

        create_default_config_file(path)
        console.print(f"[green]Created config file: {path}[/green]")

    elif action == "paths":
        from finlab_sentinel.config.loader import CONFIG_SEARCH_PATHS

        console.print("Config file search paths:")
        for path in CONFIG_SEARCH_PATHS:
            expanded = path.expanduser()
            if expanded.exists():
                exists = "[green]exists[/green]"
            else:
                exists = "[dim]not found[/dim]"
            console.print(f"  {path} -> {exists}")

    else:
        console.print(f"[red]Unknown action: {action}[/red]")
        raise typer.Exit(1)


@app.command("info")
def info() -> None:
    """Show storage information and statistics."""
    from finlab_sentinel.config.loader import load_config
    from finlab_sentinel.storage.cleanup import get_storage_info

    config = load_config(_config_path)
    stats = get_storage_info(config)

    console.print("[bold]finlab-sentinel Storage Info[/bold]\n")
    console.print(f"Storage Path: {stats['storage_path']}")
    console.print(f"Total Backups: {stats['total_backups']}")
    console.print(f"Unique Datasets: {stats['unique_datasets']}")
    console.print(f"Total Size: {_format_size(stats['total_size_bytes'])}")

    if stats["oldest"]:
        console.print(f"Oldest Backup: {stats['oldest']}")
    if stats["newest"]:
        console.print(f"Newest Backup: {stats['newest']}")


def _format_size(size_bytes: int) -> str:
    """Format bytes as human-readable size."""
    size = float(size_bytes)
    for unit in ["B", "KB", "MB", "GB"]:
        if size < 1024:
            return f"{size:.1f} {unit}"
        size /= 1024
    return f"{size:.1f} TB"


if __name__ == "__main__":
    app()



"""Configuration module for finlab-sentinel."""

from finlab_sentinel.config.loader import load_config
from finlab_sentinel.config.schema import (
    AnomalyBehavior,
    AnomalyConfig,
    ComparisonConfig,
    ComparisonPoliciesConfig,
    LoggingConfig,
    PolicyMode,
    SentinelConfig,
    StorageConfig,
)

__all__ = [
    "load_config",
    "SentinelConfig",
    "StorageConfig",
    "ComparisonConfig",
    "ComparisonPoliciesConfig",
    "AnomalyConfig",
    "LoggingConfig",
    "AnomalyBehavior",
    "PolicyMode",
]



"""Storage module for finlab-sentinel."""

from finlab_sentinel.storage.backend import BackupMetadata, StorageBackend
from finlab_sentinel.storage.parquet import ParquetStorage

__all__ = [
    "StorageBackend",
    "BackupMetadata",
    "ParquetStorage",
]



"""Tests for CLI commands."""

from pathlib import Path
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
from typer.testing import CliRunner

from finlab_sentinel.cli.main import app
from finlab_sentinel.config.schema import SentinelConfig, StorageConfig
from finlab_sentinel.storage.parquet import ParquetStorage

runner = CliRunner()


@pytest.fixture
def mock_config(tmp_path: Path) -> SentinelConfig:
    """Create mock config for CLI tests."""
    return SentinelConfig(storage=StorageConfig(path=tmp_path))


@pytest.fixture
def populated_storage(tmp_path: Path, sample_df: pd.DataFrame) -> ParquetStorage:
    """Create storage with some test data."""
    storage = ParquetStorage(base_path=tmp_path / "backups")
    storage.save("test__dataset1", "test:dataset1", sample_df, "hash1")
    storage.save("test__dataset2", "test:dataset2", sample_df, "hash2")
    return storage


class TestListCommand:
    """Tests for list command."""

    def test_list_empty_storage(self, mock_config: SentinelConfig):
        """Verify list with no backups."""
        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["list"])

        assert result.exit_code == 0
        assert "No backups found" in result.stdout

    def test_list_with_backups(
        self, mock_config: SentinelConfig, sample_df: pd.DataFrame
    ):
        """Verify list shows existing backups."""
        # Create some backups
        storage = ParquetStorage(
            base_path=mock_config.get_storage_path(),
            compression=mock_config.storage.compression,
        )
        storage.save("test__ds", "test:ds", sample_df, "hash123")

        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["list"])

        assert result.exit_code == 0
        assert "test:ds" in result.stdout

    def test_list_json_format(
        self, mock_config: SentinelConfig, sample_df: pd.DataFrame
    ):
        """Verify list with JSON format."""
        storage = ParquetStorage(
            base_path=mock_config.get_storage_path(),
            compression=mock_config.storage.compression,
        )
        storage.save("test__ds", "test:ds", sample_df, "hash123")

        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["list", "--format", "json"])

        assert result.exit_code == 0
        assert "[" in result.stdout  # JSON array

    def test_list_csv_format(
        self, mock_config: SentinelConfig, sample_df: pd.DataFrame
    ):
        """Verify list with CSV format."""
        storage = ParquetStorage(
            base_path=mock_config.get_storage_path(),
            compression=mock_config.storage.compression,
        )
        storage.save("test__ds", "test:ds", sample_df, "hash123")

        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["list", "--format", "csv"])

        assert result.exit_code == 0
        assert "dataset,backup_key" in result.stdout


class TestCleanupCommand:
    """Tests for cleanup command."""

    def test_cleanup_no_expired(self, mock_config: SentinelConfig):
        """Verify cleanup with no expired backups."""
        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["cleanup"])

        assert result.exit_code == 0
        assert "No backups older than" in result.stdout

    def test_cleanup_dry_run(
        self, mock_config: SentinelConfig, sample_df: pd.DataFrame
    ):
        """Verify cleanup dry run doesn't delete."""
        storage = ParquetStorage(
            base_path=mock_config.get_storage_path(),
            compression=mock_config.storage.compression,
        )
        storage.save("test__ds", "test:ds", sample_df, "hash123")

        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            # Use 0 days to make backup "expired"
            result = runner.invoke(app, ["cleanup", "--days", "0", "--dry-run"])

        assert result.exit_code == 0
        assert "Dry run" in result.stdout

        # Backup should still exist
        assert storage.load_latest("test__ds") is not None


class TestExportCommand:
    """Tests for export command."""

    def test_export_not_found(self, mock_config: SentinelConfig):
        """Verify export error for missing dataset."""
        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["export", "missing:dataset"])

        assert result.exit_code == 1
        assert "No backup found" in result.stdout

    def test_export_to_parquet(
        self, mock_config: SentinelConfig, sample_df: pd.DataFrame, tmp_path: Path
    ):
        """Verify export to parquet."""
        storage = ParquetStorage(
            base_path=mock_config.get_storage_path(),
            compression=mock_config.storage.compression,
        )
        storage.save("test__ds", "test:ds", sample_df, "hash123")

        output_dir = tmp_path / "output"
        output_dir.mkdir()

        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(
                app, ["export", "test:ds", "--output", str(output_dir)]
            )

        assert result.exit_code == 0
        assert "Exported to" in result.stdout

    def test_export_to_csv(
        self, mock_config: SentinelConfig, sample_df: pd.DataFrame, tmp_path: Path
    ):
        """Verify export to CSV."""
        storage = ParquetStorage(
            base_path=mock_config.get_storage_path(),
            compression=mock_config.storage.compression,
        )
        storage.save("test__ds", "test:ds", sample_df, "hash123")

        output_dir = tmp_path / "output"
        output_dir.mkdir()

        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(
                app,
                ["export", "test:ds", "--output", str(output_dir), "--format", "csv"],
            )

        assert result.exit_code == 0
        # Check CSV file was created
        csv_files = list(output_dir.glob("*.csv"))
        assert len(csv_files) == 1


class TestConfigCommand:
    """Tests for config command."""

    def test_config_show(self, mock_config: SentinelConfig):
        """Verify config show displays config."""
        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["config", "show"])

        assert result.exit_code == 0
        assert "storage" in result.stdout

    def test_config_validate_success(self, mock_config: SentinelConfig):
        """Verify config validate passes for valid config."""
        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["config", "validate"])

        assert result.exit_code == 0
        assert "valid" in result.stdout.lower()

    def test_config_init(self, tmp_path: Path):
        """Verify config init creates file."""
        output_file = tmp_path / "new_sentinel.toml"

        result = runner.invoke(app, ["config", "init", "--output", str(output_file)])

        assert result.exit_code == 0
        assert output_file.exists()
        assert "Created config file" in result.stdout

    def test_config_paths(self):
        """Verify config paths lists search paths."""
        result = runner.invoke(app, ["config", "paths"])

        assert result.exit_code == 0
        assert "sentinel.toml" in result.stdout

    def test_config_unknown_action(self):
        """Verify error for unknown action."""
        result = runner.invoke(app, ["config", "unknown"])

        assert result.exit_code == 1
        assert "Unknown action" in result.stdout


class TestInfoCommand:
    """Tests for info command."""

    def test_info_shows_stats(
        self, mock_config: SentinelConfig, sample_df: pd.DataFrame
    ):
        """Verify info shows storage stats."""
        storage = ParquetStorage(
            base_path=mock_config.get_storage_path(),
            compression=mock_config.storage.compression,
        )
        storage.save("test__ds", "test:ds", sample_df, "hash123")

        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["info"])

        assert result.exit_code == 0
        assert "Storage Path" in result.stdout
        assert "Total Backups" in result.stdout


class TestAcceptCommand:
    """Tests for accept command."""

    def test_accept_missing_dataset(self, mock_config: SentinelConfig):
        """Verify accept fails for missing dataset."""
        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            with patch(
                "finlab_sentinel.core.interceptor.accept_current_data",
                return_value=False,
            ):
                result = runner.invoke(app, ["accept", "missing:dataset"])

        assert result.exit_code == 1
        assert "Failed to accept" in result.stdout

    def test_accept_success(self, mock_config: SentinelConfig):
        """Verify accept succeeds."""
        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            with patch(
                "finlab_sentinel.core.interceptor.accept_current_data",
                return_value=True,
            ):
                result = runner.invoke(
                    app, ["accept", "test:dataset", "--reason", "Test reason"]
                )

        assert result.exit_code == 0
        assert "Accepted new data" in result.stdout


class TestDiffCommand:
    """Tests for diff command."""

    def test_diff_no_backup(self, mock_config: SentinelConfig):
        """Verify diff fails when no backup exists."""
        with patch(
            "finlab_sentinel.config.loader.load_config", return_value=mock_config
        ):
            result = runner.invoke(app, ["diff", "missing:dataset"])

        assert result.exit_code == 1
        assert "No backup found" in result.stdout

    def test_diff_identical(self, mock_config: SentinelConfig, sample_df: pd.DataFrame):
        """Verify diff shows no differences for identical data."""
        storage = ParquetStorage(
            base_path=mock_config.get_storage_path(),
            compression=mock_config.storage.compression,
        )
        storage.save("test__ds", "test:ds", sample_df, "hash123")

        # Mock finlab.data.get to return same data
        import sys
        from types import ModuleType

        mock_data = MagicMock()
        mock_data.get = MagicMock(return_value=sample_df.copy())

        mock_finlab = ModuleType("finlab")
        mock_finlab.data = mock_data

        sys.modules["finlab"] = mock_finlab

        try:
            with patch(
                "finlab_sentinel.config.loader.load_config", return_value=mock_config
            ):
                result = runner.invoke(app, ["diff", "test:ds"])

            assert result.exit_code == 0
            assert "No differences" in result.stdout
        finally:
            if "finlab" in sys.modules:
                del sys.modules["finlab"]


class TestFormatSize:
    """Tests for _format_size helper."""

    def test_bytes(self):
        """Verify bytes formatting."""
        from finlab_sentinel.cli.main import _format_size

        assert _format_size(500) == "500.0 B"

    def test_kilobytes(self):
        """Verify KB formatting."""
        from finlab_sentinel.cli.main import _format_size

        assert _format_size(1500) == "1.5 KB"

    def test_megabytes(self):
        """Verify MB formatting."""
        from finlab_sentinel.cli.main import _format_size

        assert _format_size(1500000) == "1.4 MB"

    def test_gigabytes(self):
        """Verify GB formatting."""
        from finlab_sentinel.cli.main import _format_size

        assert _format_size(1500000000) == "1.4 GB"

    def test_terabytes(self):
        """Verify TB formatting."""
        from finlab_sentinel.cli.main import _format_size

        assert _format_size(1500000000000) == "1.4 TB"



"""Registry for storing original functions."""

from __future__ import annotations

from collections.abc import Callable

# Global registry for original functions
_original_functions: dict[str, Callable] = {}

# Global flag for enabled state
_enabled: bool = False


def store_original(name: str, func: Callable) -> None:
    """Store original function in registry.

    Args:
        name: Name/key for the function
        func: Original function to store
    """
    _original_functions[name] = func


def get_original(name: str) -> Callable | None:
    """Get original function from registry.

    Args:
        name: Name/key for the function

    Returns:
        Original function or None if not found
    """
    return _original_functions.get(name)


def remove_original(name: str) -> Callable | None:
    """Remove and return original function from registry.

    Args:
        name: Name/key for the function

    Returns:
        Removed function or None if not found
    """
    return _original_functions.pop(name, None)


def clear_registry() -> None:
    """Clear all stored functions."""
    _original_functions.clear()


def set_enabled(enabled: bool) -> None:
    """Set enabled state.

    Args:
        enabled: Whether sentinel is enabled
    """
    global _enabled
    _enabled = enabled


def get_enabled() -> bool:
    """Get enabled state.

    Returns:
        Whether sentinel is enabled
    """
    return _enabled



"""finlab-sentinel: Defensive monitoring layer for finlab data.get API."""

from finlab_sentinel.core.hooks import (
    clear_preprocess_hooks,
    register_preprocess_hook,
    unregister_preprocess_hook,
)
from finlab_sentinel.core.patcher import disable, enable, is_enabled
from finlab_sentinel.exceptions import DataAnomalyError, SentinelError

__version__ = "0.1.5"

__all__ = [
    "__version__",
    "enable",
    "disable",
    "is_enabled",
    "SentinelError",
    "DataAnomalyError",
    "register_preprocess_hook",
    "unregister_preprocess_hook",
    "clear_preprocess_hooks",
]



"""Tests for anomaly report generation and storage."""

from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from finlab_sentinel.comparison.differ import (
    CellChange,
    ChangeType,
    ComparisonResult,
    DtypeChange,
)
from finlab_sentinel.comparison.report import (
    AnomalyReport,
    _serialize_value,
    list_reports,
)


@pytest.fixture
def sample_comparison_result() -> ComparisonResult:
    """Create sample comparison result for testing."""
    return ComparisonResult(
        is_identical=False,
        old_shape=(10, 4),
        new_shape=(9, 4),
        added_rows={"2025-01-11"},
        deleted_rows={"2025-01-01"},
        added_columns=set(),
        deleted_columns=set(),
        modified_cells=[
            CellChange(
                row="2025-01-05",
                column="2330",
                old_value=100.0,
                new_value=101.0,
                change_type=ChangeType.VALUE_MODIFIED,
            )
        ],
        dtype_changes=[DtypeChange("col1", "float64", "float32")],
        na_type_changes=[
            CellChange(
                row="2025-01-02",
                column="2331",
                old_value=np.nan,
                new_value=pd.NA,
                change_type=ChangeType.NA_TYPE_CHANGED,
            )
        ],
    )


@pytest.fixture
def sample_report(sample_comparison_result: ComparisonResult) -> AnomalyReport:
    """Create sample anomaly report."""
    return AnomalyReport(
        dataset="price:收盤價",
        backup_key="price__收盤價",
        detected_at=datetime(2025, 1, 10, 12, 30, 0),
        comparison_result=sample_comparison_result,
        policy_name="append_only",
        violation_message="Deleted rows detected",
        old_hash="abc123",
        new_hash="def456",
    )


class TestAnomalyReport:
    """Tests for AnomalyReport class."""

    def test_summary_property(self, sample_report: AnomalyReport):
        """Verify summary includes key info."""
        summary = sample_report.summary

        assert "price:收盤價" in summary
        assert "append_only violated" in summary

    def test_to_dict(self, sample_report: AnomalyReport):
        """Verify to_dict serialization."""
        data = sample_report.to_dict()

        assert data["dataset"] == "price:收盤價"
        assert data["backup_key"] == "price__收盤價"
        assert data["policy_name"] == "append_only"
        assert data["old_hash"] == "abc123"
        assert data["new_hash"] == "def456"
        assert data["comparison"]["is_identical"] is False
        assert data["comparison"]["old_shape"] == [10, 4]
        assert len(data["comparison"]["modified_cells"]) == 1
        assert len(data["comparison"]["dtype_changes"]) == 1
        assert len(data["comparison"]["na_type_changes"]) == 1

    def test_to_json(self, sample_report: AnomalyReport):
        """Verify JSON serialization."""
        json_str = sample_report.to_json()

        assert isinstance(json_str, str)
        assert "price:收盤價" in json_str
        assert "append_only" in json_str

    def test_save_and_load(self, sample_report: AnomalyReport, tmp_path: Path):
        """Verify save and load roundtrip."""
        # Save
        filepath = sample_report.save(tmp_path)

        assert filepath.exists()
        assert "anomaly.json" in filepath.name

        # Load
        loaded = AnomalyReport.load(filepath)

        assert loaded.dataset == sample_report.dataset
        assert loaded.backup_key == sample_report.backup_key
        assert loaded.policy_name == sample_report.policy_name
        assert loaded.old_hash == sample_report.old_hash

    def test_from_dict(self, sample_report: AnomalyReport):
        """Verify from_dict deserialization."""
        data = sample_report.to_dict()
        loaded = AnomalyReport.from_dict(data)

        assert loaded.dataset == sample_report.dataset
        assert (
            loaded.comparison_result.is_identical
            == sample_report.comparison_result.is_identical
        )
        assert len(loaded.comparison_result.modified_cells) == len(
            sample_report.comparison_result.modified_cells
        )

    def test_save_creates_directory(self, sample_report: AnomalyReport, tmp_path: Path):
        """Verify save creates directory if needed."""
        nested_path = tmp_path / "nested" / "reports"

        filepath = sample_report.save(nested_path)

        assert filepath.exists()
        assert nested_path.exists()


class TestListReports:
    """Tests for list_reports function."""

    def test_empty_directory(self, tmp_path: Path):
        """Verify empty list for empty directory."""
        reports = list_reports(tmp_path)
        assert reports == []

    def test_nonexistent_directory(self, tmp_path: Path):
        """Verify empty list for nonexistent directory."""
        nonexistent = tmp_path / "does_not_exist"
        reports = list_reports(nonexistent)
        assert reports == []

    def test_lists_reports(self, sample_report: AnomalyReport, tmp_path: Path):
        """Verify reports are listed."""
        # Save multiple reports
        sample_report.save(tmp_path)

        # Change timestamp and save again
        sample_report.detected_at = datetime(2025, 1, 11, 12, 0, 0)
        sample_report.save(tmp_path)

        reports = list_reports(tmp_path)

        assert len(reports) == 2
        # Should be sorted newest first
        assert "2025-01-11" in reports[0].name


class TestSerializeValue:
    """Tests for _serialize_value function."""

    def test_serialize_none(self):
        """Verify None serialization."""
        result = _serialize_value(None)
        assert result == {"__na__": "None"}

    def test_serialize_pd_na(self):
        """Verify pd.NA serialization."""
        result = _serialize_value(pd.NA)
        assert result == {"__na__": "pd.NA"}

    def test_serialize_np_nan(self):
        """Verify np.nan serialization."""
        result = _serialize_value(np.nan)
        assert result == {"__na__": "np.nan"}

    def test_serialize_numpy_int(self):
        """Verify numpy int serialization."""
        result = _serialize_value(np.int64(42))
        assert result == 42.0

    def test_serialize_numpy_float(self):
        """Verify numpy float serialization."""
        result = _serialize_value(np.float64(3.14))
        assert result == 3.14

    def test_serialize_numpy_array(self):
        """Verify numpy array serialization."""
        result = _serialize_value(np.array([1, 2, 3]))
        assert result == [1, 2, 3]

    def test_serialize_datetime(self):
        """Verify datetime serialization."""
        dt = datetime(2025, 1, 10, 12, 30, 0)
        result = _serialize_value(dt)
        assert result == "2025-01-10T12:30:00"

    def test_serialize_timestamp(self):
        """Verify pandas Timestamp serialization."""
        ts = pd.Timestamp("2025-01-10 12:30:00")
        result = _serialize_value(ts)
        assert "2025-01-10" in result

    def test_serialize_string(self):
        """Verify string serialization."""
        result = _serialize_value("test")
        assert result == "test"

    def test_serialize_int(self):
        """Verify int serialization."""
        result = _serialize_value(42)
        assert result == "42"



"""Tests for storage backend."""

import time
from datetime import datetime, timedelta

import pandas as pd

from finlab_sentinel.storage.parquet import ParquetStorage, sanitize_backup_key


class TestSanitizeBackupKey:
    """Tests for backup key sanitization."""

    def test_basic_dataset_name(self):
        """Verify basic dataset name is sanitized."""
        key = sanitize_backup_key("price:收盤價")
        assert key == "price__收盤價"

    def test_with_universe_hash(self):
        """Verify universe hash is appended."""
        key = sanitize_backup_key("price:收盤價", "abc123")
        assert key == "price__收盤價__universe_abc123"

    def test_removes_slashes(self):
        """Verify slashes are removed."""
        key = sanitize_backup_key("path/to/data")
        assert "/" not in key


class TestParquetStorage:
    """Tests for ParquetStorage class."""

    def test_save_creates_file(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify save creates parquet file."""
        metadata = parquet_storage.save(
            backup_key="test_dataset",
            dataset="test:dataset",
            data=sample_df,
            content_hash="abc123",
        )

        assert metadata.file_path.exists()
        assert metadata.row_count == len(sample_df)
        assert metadata.column_count == len(sample_df.columns)

    def test_load_returns_saved_data(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify loaded DataFrame matches saved."""
        parquet_storage.save(
            backup_key="test_dataset",
            dataset="test:dataset",
            data=sample_df,
            content_hash="abc123",
        )

        result = parquet_storage.load_latest("test_dataset")

        assert result is not None
        loaded_df, metadata = result
        # Check values match (ignore index freq attribute which may not be preserved)
        pd.testing.assert_frame_equal(
            loaded_df.reset_index(drop=True),
            sample_df.reset_index(drop=True),
        )
        # Check index values match
        assert list(loaded_df.index) == list(sample_df.index)

    def test_load_preserves_dtypes(self, parquet_storage: ParquetStorage):
        """Verify dtypes are preserved through save/load."""
        df = pd.DataFrame(
            {
                "int_col": [1, 2, 3],
                "float_col": [1.1, 2.2, 3.3],
                "str_col": ["a", "b", "c"],
            }
        )

        parquet_storage.save(
            backup_key="dtype_test",
            dataset="test",
            data=df,
            content_hash="abc",
        )

        result = parquet_storage.load_latest("dtype_test")
        loaded_df, _ = result

        assert loaded_df["int_col"].dtype == df["int_col"].dtype
        assert loaded_df["float_col"].dtype == df["float_col"].dtype

    def test_list_backups(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify all backups are listed."""
        # Save multiple backups
        parquet_storage.save("ds1", "dataset1", sample_df, "hash1")
        parquet_storage.save("ds2", "dataset2", sample_df, "hash2")

        backups = parquet_storage.list_backups()

        assert len(backups) >= 2

    def test_list_backups_filtered(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify backups can be filtered by key."""
        parquet_storage.save("ds1", "dataset1", sample_df, "hash1")
        parquet_storage.save("ds2", "dataset2", sample_df, "hash2")

        backups = parquet_storage.list_backups("ds1")

        assert len(backups) == 1
        assert backups[0].backup_key == "ds1"

    def test_load_nonexistent_returns_none(self, parquet_storage: ParquetStorage):
        """Verify loading nonexistent backup returns None."""
        result = parquet_storage.load_latest("nonexistent")
        assert result is None

    def test_get_latest_metadata(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify getting metadata without loading data."""
        parquet_storage.save("test", "test", sample_df, "hash1")

        metadata = parquet_storage.get_latest_metadata("test")

        assert metadata is not None
        assert metadata.content_hash == "hash1"

    def test_delete_backup(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify backup deletion."""
        parquet_storage.save("to_delete", "test", sample_df, "hash")

        deleted = parquet_storage.delete("to_delete")

        assert deleted == 1
        assert parquet_storage.load_latest("to_delete") is None

    def test_accept_new_data(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify accept_new_data saves with reason."""
        metadata = parquet_storage.accept_new_data(
            backup_key="accepted",
            data=sample_df,
            content_hash="new_hash",
            dataset="test",
            reason="Data correction confirmed",
        )

        assert metadata is not None
        assert metadata.content_hash == "new_hash"

    def test_get_stats(self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame):
        """Verify storage statistics."""
        parquet_storage.save("stat_test", "test", sample_df, "hash")

        stats = parquet_storage.get_stats()

        assert stats["total_backups"] >= 1
        assert stats["total_size_bytes"] > 0

    def test_backup_file_uses_minute_timestamp(self, parquet_storage: ParquetStorage):
        """Verify backup filename uses minute-level timestamp."""
        dt = datetime(2024, 1, 4, 9, 30, 45)
        path = parquet_storage._get_backup_file("test_key", dt)
        assert path.name == "2024-01-04T09-30.parquet"

    def test_backup_file_different_minutes(self, parquet_storage: ParquetStorage):
        """Verify different minutes produce different filenames."""
        dt1 = datetime(2024, 1, 4, 9, 30, 0)
        dt2 = datetime(2024, 1, 4, 9, 31, 0)
        path1 = parquet_storage._get_backup_file("test_key", dt1)
        path2 = parquet_storage._get_backup_file("test_key", dt2)
        assert path1.name != path2.name
        assert path1.name == "2024-01-04T09-30.parquet"
        assert path2.name == "2024-01-04T09-31.parquet"

    def test_multiple_backups_same_day(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify multiple backups on the same day are not overwritten."""
        # Save first backup
        parquet_storage.save("multi_test", "test", sample_df, "hash1")
        # Wait a bit to ensure different timestamp
        time.sleep(0.1)
        # Modify and save again
        modified_df = sample_df.copy()
        modified_df.iloc[0, 0] = 999.99
        parquet_storage.save("multi_test", "test", modified_df, "hash2")

        # Both backups should exist
        backups = parquet_storage.list_backups("multi_test")
        assert len(backups) == 2

    def test_cleanup_expired_keeps_min_per_key(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify cleanup keeps minimum backups per key even if all expired."""
        # Create 5 backups with different old timestamps (spaced by minutes)
        # This ensures different filenames
        base_time = datetime.now() - timedelta(days=30)
        for i in range(5):
            parquet_storage.save("cleanup_test", "test", sample_df, f"hash{i}")
            # Set different old timestamps (each 10 minutes apart) to get different filenames
            with parquet_storage.index._connect() as conn:
                old_date = (base_time - timedelta(minutes=i * 10)).isoformat()
                conn.execute(
                    "UPDATE backups SET created_at = ? WHERE content_hash = ?",
                    (old_date, f"hash{i}"),
                )

        # All 5 are older than 7 days, but should keep 3 (in index)
        parquet_storage.cleanup_expired(retention_days=7, min_keep_per_key=3)

        # Index should have 3 remaining entries
        remaining = parquet_storage.list_backups("cleanup_test")
        assert len(remaining) == 3

    def test_cleanup_expired_respects_retention_when_enough_recent(
        self, parquet_storage: ParquetStorage, sample_df: pd.DataFrame
    ):
        """Verify cleanup deletes old backups when enough recent ones exist."""
        # Create 3 old backups with different timestamps
        base_old_time = datetime.now() - timedelta(days=30)
        for i in range(3):
            parquet_storage.save("retention_test", "test", sample_df, f"old_hash{i}")
            with parquet_storage.index._connect() as conn:
                old_date = (base_old_time - timedelta(minutes=i * 10)).isoformat()
                conn.execute(
                    "UPDATE backups SET created_at = ? WHERE content_hash = ?",
                    (old_date, f"old_hash{i}"),
                )

        # Create 3 recent backups with different timestamps
        base_new_time = datetime.now()
        for i in range(3):
            parquet_storage.save("retention_test", "test", sample_df, f"new_hash{i}")
            with parquet_storage.index._connect() as conn:
                new_date = (base_new_time - timedelta(minutes=i * 10)).isoformat()
                conn.execute(
                    "UPDATE backups SET created_at = ? WHERE content_hash = ?",
                    (new_date, f"new_hash{i}"),
                )

        # Should delete the 3 old ones, keep the 3 recent ones
        parquet_storage.cleanup_expired(retention_days=7, min_keep_per_key=3)

        # Index should have 3 remaining entries (the recent ones)
        remaining = parquet_storage.list_backups("retention_test")
        assert len(remaining) == 3
        # All remaining should be recent (new_hash*)
        for backup in remaining:
            assert backup.content_hash.startswith("new_hash")



"""Monkey patching logic for finlab data.get."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from finlab_sentinel.core.registry import (
    get_enabled,
    get_original,
    remove_original,
    set_enabled,
    store_original,
)
from finlab_sentinel.exceptions import (
    SentinelAlreadyEnabledError,
    SentinelNotEnabledError,
)

if TYPE_CHECKING:
    from finlab_sentinel.config.schema import SentinelConfig

logger = logging.getLogger(__name__)


def enable(config: SentinelConfig | None = None) -> None:
    """Enable sentinel monitoring by monkey patching finlab.data.get.

    Args:
        config: Optional configuration. If None, loads from sentinel.toml
                or uses defaults.

    Raises:
        ImportError: If finlab package is not installed.
        SentinelAlreadyEnabledError: If sentinel is already enabled.
    """
    if is_enabled():
        raise SentinelAlreadyEnabledError(
            "Sentinel is already enabled. Call disable() first."
        )

    # Load config if not provided
    if config is None:
        from finlab_sentinel.config.loader import load_config

        config = load_config()

    # Configure logging
    _configure_logging(config)

    # Import finlab
    try:
        from finlab import data
    except ImportError as e:
        raise ImportError(
            "finlab package is not installed. Install it with: pip install finlab"
        ) from e

    # Store original function
    store_original("data.get", data.get)

    # Create interceptor
    from finlab_sentinel.core.interceptor import DataInterceptor

    interceptor = DataInterceptor(data.get, config)

    # Monkey patch
    data.get = interceptor

    # Mark as enabled
    set_enabled(True)

    # Cleanup expired backups on startup
    from finlab_sentinel.storage.cleanup import cleanup_on_startup

    cleanup_on_startup(config)

    logger.info("finlab-sentinel enabled")


def disable() -> None:
    """Disable sentinel monitoring and restore original data.get.

    Raises:
        SentinelNotEnabledError: If sentinel is not currently enabled.
    """
    if not is_enabled():
        raise SentinelNotEnabledError("Sentinel is not currently enabled.")

    # Get original function
    original_get = get_original("data.get")

    if original_get is None:
        logger.warning("Original data.get not found in registry")
    else:
        # Restore original
        try:
            from finlab import data

            data.get = original_get
        except ImportError:
            pass  # finlab not available, nothing to restore

    # Clean up
    remove_original("data.get")
    set_enabled(False)

    logger.info("finlab-sentinel disabled")


def is_enabled() -> bool:
    """Check if sentinel monitoring is currently enabled.

    Returns:
        True if enabled, False otherwise
    """
    return get_enabled()


def _configure_logging(config: SentinelConfig) -> None:
    """Configure logging based on config.

    Args:
        config: Sentinel configuration
    """
    sentinel_logger = logging.getLogger("finlab_sentinel")

    # Set level
    level = getattr(logging, config.logging.level)
    sentinel_logger.setLevel(level)

    # Add file handler if configured
    if config.logging.file:
        file_path = config.logging.file
        file_path.parent.mkdir(parents=True, exist_ok=True)

        file_handler = logging.FileHandler(file_path)
        file_handler.setLevel(level)
        file_handler.setFormatter(
            logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        )
        sentinel_logger.addHandler(file_handler)

    # Ensure there's at least a console handler if no handlers exist
    if not sentinel_logger.handlers:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        console_handler.setFormatter(
            logging.Formatter("%(levelname)s - %(name)s - %(message)s")
        )
        sentinel_logger.addHandler(console_handler)



"""Preprocess hooks for data comparison."""

from __future__ import annotations

import fnmatch
import logging
from collections.abc import Callable
from typing import TypeAlias

import pandas as pd

logger = logging.getLogger(__name__)

# Type alias for preprocess function
PreprocessFn: TypeAlias = Callable[[pd.DataFrame], pd.DataFrame]


class PreprocessHookRegistry:
    """Registry for dataset-specific preprocess hooks.

    Preprocess hooks are applied before comparison to normalize data.
    This is useful for ignoring expected variations (e.g., rounding,
    column ordering) that shouldn't trigger anomaly detection.

    The original data is always returned to the user; preprocess is
    only used for comparison and backup storage.
    """

    def __init__(self) -> None:
        """Initialize empty registry."""
        # Exact match hooks: dataset -> function
        self._exact_hooks: dict[str, PreprocessFn] = {}
        # Pattern hooks: pattern -> function (supports wildcards)
        self._pattern_hooks: dict[str, PreprocessFn] = {}

    def register(
        self,
        dataset: str,
        fn: PreprocessFn,
    ) -> None:
        """Register a preprocess hook for a dataset.

        Args:
            dataset: Dataset name or pattern (supports * and ? wildcards)
            fn: Function that takes DataFrame and returns preprocessed DataFrame

        Examples:
            >>> registry.register("price:收盤價", lambda df: df.round(2))
            >>> registry.register("price:*", my_price_preprocessor)
            >>> registry.register("fundamental_features:*", my_fundamental_preprocessor)
        """
        if "*" in dataset or "?" in dataset:
            self._pattern_hooks[dataset] = fn
            logger.debug(f"Registered pattern preprocess hook: {dataset}")
        else:
            self._exact_hooks[dataset] = fn
            logger.debug(f"Registered exact preprocess hook: {dataset}")

    def unregister(self, dataset: str) -> bool:
        """Unregister a preprocess hook.

        Args:
            dataset: Dataset name or pattern to unregister

        Returns:
            True if hook was found and removed, False otherwise
        """
        if dataset in self._exact_hooks:
            del self._exact_hooks[dataset]
            logger.debug(f"Unregistered exact preprocess hook: {dataset}")
            return True
        if dataset in self._pattern_hooks:
            del self._pattern_hooks[dataset]
            logger.debug(f"Unregistered pattern preprocess hook: {dataset}")
            return True
        return False

    def get(self, dataset: str) -> PreprocessFn | None:
        """Get preprocess hook for a dataset.

        Exact match takes precedence over pattern match.
        If multiple patterns match, the first registered one is used.

        Args:
            dataset: Dataset name

        Returns:
            Preprocess function or None if no hook registered
        """
        # Exact match first
        if dataset in self._exact_hooks:
            return self._exact_hooks[dataset]

        # Pattern match
        for pattern, fn in self._pattern_hooks.items():
            if fnmatch.fnmatch(dataset, pattern):
                return fn

        return None

    def apply(self, dataset: str, df: pd.DataFrame) -> pd.DataFrame:
        """Apply preprocess hook to DataFrame if one exists.

        Args:
            dataset: Dataset name
            df: DataFrame to preprocess

        Returns:
            Preprocessed DataFrame, or original if no hook registered
        """
        fn = self.get(dataset)
        if fn is None:
            return df

        try:
            result = fn(df)
            if not isinstance(result, pd.DataFrame):
                logger.warning(
                    f"Preprocess hook for {dataset} did not return DataFrame, "
                    f"got {type(result).__name__}. Using original data."
                )
                return df
            logger.debug(f"Applied preprocess hook for {dataset}")
            return result
        except Exception as e:
            logger.error(
                f"Preprocess hook failed for {dataset}: {e}. Using original data."
            )
            return df

    def clear(self) -> None:
        """Clear all registered hooks."""
        self._exact_hooks.clear()
        self._pattern_hooks.clear()
        logger.debug("Cleared all preprocess hooks")

    def list_hooks(self) -> dict[str, list[str]]:
        """List all registered hooks.

        Returns:
            Dict with 'exact' and 'pattern' keys containing dataset names
        """
        return {
            "exact": list(self._exact_hooks.keys()),
            "pattern": list(self._pattern_hooks.keys()),
        }


# Global registry instance
_registry = PreprocessHookRegistry()


def get_registry() -> PreprocessHookRegistry:
    """Get the global preprocess hook registry.

    Returns:
        Global PreprocessHookRegistry instance
    """
    return _registry


def register_preprocess_hook(dataset: str, fn: PreprocessFn) -> None:
    """Register a preprocess hook for a dataset.

    Convenience function that uses the global registry.

    Args:
        dataset: Dataset name or pattern (supports * and ? wildcards)
        fn: Function that takes DataFrame and returns preprocessed DataFrame

    Example:
        >>> import finlab_sentinel
        >>> finlab_sentinel.register_preprocess_hook(
        ...     "price:收盤價",
        ...     lambda df: df.round(2)
        ... )
    """
    _registry.register(dataset, fn)


def unregister_preprocess_hook(dataset: str) -> bool:
    """Unregister a preprocess hook.

    Args:
        dataset: Dataset name or pattern to unregister

    Returns:
        True if hook was found and removed, False otherwise
    """
    return _registry.unregister(dataset)


def clear_preprocess_hooks() -> None:
    """Clear all registered preprocess hooks."""
    _registry.clear()



"""DataFrame difference detection."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any

import numpy as np
import pandas as pd


class ChangeType(str, Enum):
    """Types of changes detected."""

    ROW_ADDED = "row_added"
    ROW_DELETED = "row_deleted"
    COLUMN_ADDED = "column_added"
    COLUMN_DELETED = "column_deleted"
    VALUE_MODIFIED = "value_modified"
    DTYPE_CHANGED = "dtype_changed"
    NA_TYPE_CHANGED = "na_type_changed"


@dataclass
class CellChange:
    """Represents a single cell value change."""

    row: Any  # Index value (e.g., date)
    column: Any  # Column name (e.g., stock symbol)
    old_value: Any
    new_value: Any
    change_type: ChangeType

    def __str__(self) -> str:
        if self.change_type == ChangeType.VALUE_MODIFIED:
            return f"[{self.row}, {self.column}]: {self.old_value} -> {self.new_value}"
        elif self.change_type == ChangeType.NA_TYPE_CHANGED:
            return (
                f"[{self.row}, {self.column}]: NA type changed "
                f"{_get_na_type(self.old_value)} -> {_get_na_type(self.new_value)}"
            )
        return f"[{self.row}, {self.column}]: {self.change_type.value}"


@dataclass
class DtypeChange:
    """Represents a dtype change for a column."""

    column: Any
    old_dtype: str
    new_dtype: str

    def __str__(self) -> str:
        return f"{self.column}: {self.old_dtype} -> {self.new_dtype}"


# Maximum number of cell changes to track individually
# Beyond this, only count is reported for performance
MAX_CELL_CHANGES = 10


@dataclass
class ComparisonResult:
    """Result of DataFrame comparison."""

    is_identical: bool
    added_rows: set[Any] = field(default_factory=set)
    deleted_rows: set[Any] = field(default_factory=set)
    added_columns: set[Any] = field(default_factory=set)
    deleted_columns: set[Any] = field(default_factory=set)
    modified_cells: list[CellChange] = field(default_factory=list)
    dtype_changes: list[DtypeChange] = field(default_factory=list)
    na_type_changes: list[CellChange] = field(default_factory=list)
    na_to_value_cells: list[CellChange] = field(default_factory=list)

    # Counts (may be > len(list) when exceeding MAX_CELL_CHANGES)
    modified_cells_count: int = 0
    na_type_changes_count: int = 0
    na_to_value_cells_count: int = 0

    # Metrics
    old_shape: tuple[int, int] = (0, 0)
    new_shape: tuple[int, int] = (0, 0)

    @property
    def total_changes(self) -> int:
        """Total number of changes detected."""
        # Use count fields which may be larger than list lengths
        modified = max(self.modified_cells_count, len(self.modified_cells))
        na_changes = max(self.na_type_changes_count, len(self.na_type_changes))
        return (
            len(self.added_rows)
            + len(self.deleted_rows)
            + len(self.added_columns)
            + len(self.deleted_columns)
            + modified
            + len(self.dtype_changes)
            + na_changes
        )

    @property
    def change_ratio(self) -> float:
        """Calculate ratio of changed cells to total cells."""
        if self.old_shape == (0, 0) and self.new_shape == (0, 0):
            return 0.0

        old_total = self.old_shape[0] * self.old_shape[1]
        new_total = self.new_shape[0] * self.new_shape[1]
        total = max(old_total, new_total, 1)

        # Calculate changes
        # Deleted rows affect all old columns
        deleted_cells = len(self.deleted_rows) * self.old_shape[1]
        # Deleted columns affect all old rows
        deleted_cells += len(self.deleted_columns) * self.old_shape[0]
        # Modified cells - use count fields which may be larger than list lengths
        modified = max(self.modified_cells_count, len(self.modified_cells))
        na_changes = max(self.na_type_changes_count, len(self.na_type_changes))

        return (deleted_cells + modified + na_changes) / total

    def is_append_only(self, ignore_na_to_value: bool = False) -> bool:
        """Check if changes are append-only (no deletions/modifications).

        Args:
            ignore_na_to_value: If True, NA→value changes are allowed

        Returns:
            True if changes are append-only
        """
        modified = max(self.modified_cells_count, len(self.modified_cells))
        na_changes = max(self.na_type_changes_count, len(self.na_type_changes))
        na_to_value = max(self.na_to_value_cells_count, len(self.na_to_value_cells))

        # If ignoring NA→value, subtract from modified count
        if ignore_na_to_value:
            modified = max(0, modified - na_to_value)

        return (
            len(self.deleted_rows) == 0
            and len(self.deleted_columns) == 0
            and modified == 0
            and len(self.dtype_changes) == 0
            and na_changes == 0
        )

    def exceeds_threshold(self, threshold: float) -> bool:
        """Check if change ratio exceeds threshold."""
        return self.change_ratio > threshold

    def summary(self) -> str:
        """Generate human-readable summary."""
        parts = []
        if self.added_rows:
            parts.append(f"+{len(self.added_rows)} rows")
        if self.deleted_rows:
            parts.append(f"-{len(self.deleted_rows)} rows")
        if self.added_columns:
            parts.append(f"+{len(self.added_columns)} columns")
        if self.deleted_columns:
            parts.append(f"-{len(self.deleted_columns)} columns")

        # Use count fields which may be larger than list lengths
        modified = max(self.modified_cells_count, len(self.modified_cells))
        na_changes = max(self.na_type_changes_count, len(self.na_type_changes))

        if modified > 0:
            parts.append(f"{modified} modified cells")
        if self.dtype_changes:
            parts.append(f"{len(self.dtype_changes)} dtype changes")
        if na_changes > 0:
            parts.append(f"{na_changes} NA type changes")

        if not parts:
            return "No changes"

        return f"{', '.join(parts)} ({self.change_ratio:.1%} change ratio)"


class DataFrameComparer:
    """Compares two DataFrames and detects changes."""

    def __init__(
        self,
        rtol: float = 1e-5,
        atol: float = 1e-8,
        check_dtype: bool = True,
        check_na_type: bool = True,
    ) -> None:
        """Initialize comparer.

        Args:
            rtol: Relative tolerance for numeric comparisons
            atol: Absolute tolerance for numeric comparisons
            check_dtype: Whether to check for dtype changes
            check_na_type: Whether to check for NA type differences
        """
        self.rtol = rtol
        self.atol = atol
        self.check_dtype = check_dtype
        self.check_na_type = check_na_type

    def compare(
        self,
        old_df: pd.DataFrame,
        new_df: pd.DataFrame,
    ) -> ComparisonResult:
        """Compare two DataFrames and return detailed differences.

        Uses vectorized numpy operations for performance on large DataFrames.

        Args:
            old_df: Previous/cached DataFrame
            new_df: New DataFrame from data source

        Returns:
            ComparisonResult with all detected changes
        """
        result = ComparisonResult(
            is_identical=True,
            old_shape=(len(old_df), len(old_df.columns)),
            new_shape=(len(new_df), len(new_df.columns)),
        )

        # Compare index (rows)
        old_index = set(old_df.index)
        new_index = set(new_df.index)

        result.added_rows = new_index - old_index
        result.deleted_rows = old_index - new_index

        # Compare columns
        old_columns = set(old_df.columns)
        new_columns = set(new_df.columns)

        result.added_columns = new_columns - old_columns
        result.deleted_columns = old_columns - new_columns

        # Get common rows and columns
        common_idx = old_df.index.intersection(new_df.index)
        common_cols = old_df.columns.intersection(new_df.columns)

        # Check for dtype changes in common columns
        if self.check_dtype:
            for col in common_cols:
                old_dtype = str(old_df[col].dtype)
                new_dtype = str(new_df[col].dtype)
                if old_dtype != new_dtype:
                    result.dtype_changes.append(DtypeChange(col, old_dtype, new_dtype))

        # Early return if no common cells to compare
        if len(common_idx) == 0 or len(common_cols) == 0:
            result.is_identical = result.total_changes == 0
            return result

        # Align DataFrames for vectorized comparison
        old_aligned = old_df.loc[common_idx, common_cols]
        new_aligned = new_df.loc[common_idx, common_cols]

        # Vectorized NA detection
        old_na = pd.isna(old_aligned)
        new_na = pd.isna(new_aligned)

        # Both NA → equal (NA type checked separately)
        both_na = old_na & new_na

        # NA→value: old was NA, new has value
        na_to_value = old_na & ~new_na

        # value→NA: old had value, new is NA (still counts as modified)
        # one_na captures both directions
        one_na = old_na ^ new_na

        # Get numpy arrays for fast comparison
        old_vals = old_aligned.values
        new_vals = new_aligned.values

        # Build comparison mask column by column (handles mixed dtypes)
        equal_mask = np.zeros(old_vals.shape, dtype=bool)

        for col_idx, col in enumerate(common_cols):
            old_col = old_vals[:, col_idx]
            new_col = new_vals[:, col_idx]
            col_dtype = old_aligned[col].dtype

            if pd.api.types.is_numeric_dtype(col_dtype):
                # Numeric: use tolerance comparison
                with np.errstate(invalid="ignore"):
                    col_equal = np.isclose(
                        old_col.astype(float),
                        new_col.astype(float),
                        rtol=self.rtol,
                        atol=self.atol,
                        equal_nan=True,
                    )
            else:
                # Non-numeric: exact equality (handle object dtype safely)
                col_equal = np.array(
                    [
                        (pd.isna(o) and pd.isna(n)) or (o == n)
                        for o, n in zip(old_col, new_col, strict=True)
                    ]
                )

            equal_mask[:, col_idx] = col_equal

        # Combine: different if one_na OR (not equal AND not both_na)
        # one_na: one is NA, one is not → always different
        # ~equal_mask & ~both_na: values differ and not both NA
        diff_mask = one_na.values | (~equal_mask & ~both_na.values)

        # Find difference positions
        diff_rows, diff_cols = np.where(diff_mask)
        diff_count = len(diff_rows)

        # Find NA→value positions
        na_to_value_rows, na_to_value_cols = np.where(na_to_value.values)
        na_to_value_count = len(na_to_value_rows)

        # Set the counts
        result.modified_cells_count = diff_count
        result.na_to_value_cells_count = na_to_value_count

        # Only create CellChange objects if within limit
        if diff_count > 0 and diff_count <= MAX_CELL_CHANGES:
            for i, j in zip(diff_rows, diff_cols, strict=True):
                row_label = common_idx[i]
                col_label = common_cols[j]
                old_val = old_aligned.iloc[i, j]
                new_val = new_aligned.iloc[i, j]
                result.modified_cells.append(
                    CellChange(
                        row=row_label,
                        column=col_label,
                        old_value=old_val,
                        new_value=new_val,
                        change_type=ChangeType.VALUE_MODIFIED,
                    )
                )

        # Track NA→value cells separately
        if na_to_value_count > 0 and na_to_value_count <= MAX_CELL_CHANGES:
            for i, j in zip(na_to_value_rows, na_to_value_cols, strict=True):
                row_label = common_idx[i]
                col_label = common_cols[j]
                old_val = old_aligned.iloc[i, j]
                new_val = new_aligned.iloc[i, j]
                result.na_to_value_cells.append(
                    CellChange(
                        row=row_label,
                        column=col_label,
                        old_value=old_val,
                        new_value=new_val,
                        change_type=ChangeType.VALUE_MODIFIED,
                    )
                )

        # Check NA type changes (only for cells where both are NA)
        if self.check_na_type:
            na_type_changes = self._check_na_type_changes_vectorized(
                old_aligned, new_aligned, both_na
            )
            result.na_type_changes_count = len(na_type_changes)
            if len(na_type_changes) <= MAX_CELL_CHANGES:
                result.na_type_changes = na_type_changes

        # Determine if identical
        result.is_identical = result.total_changes == 0

        return result

    def _check_na_type_changes_vectorized(
        self,
        old_aligned: pd.DataFrame,
        new_aligned: pd.DataFrame,
        both_na: pd.DataFrame,
    ) -> list[CellChange]:
        """Check for NA type differences in cells where both are NA.

        For non-object columns with matching dtypes, all NAs are the same type
        (e.g., float64 always uses np.nan, Float64 always uses pd.NA).
        NA type differences can only occur in object columns or when dtypes differ.
        Since dtype differences are already detected separately, we only check
        object columns here for performance.

        Args:
            old_aligned: Aligned old DataFrame
            new_aligned: Aligned new DataFrame
            both_na: Boolean mask of cells where both are NA

        Returns:
            List of CellChange for NA type changes
        """
        if not both_na.any().any():
            return []

        # Only object columns can have mixed NA types (None, pd.NA, np.nan)
        # Non-object columns always have consistent NA type based on dtype
        object_col_indices = set()
        for idx, col in enumerate(old_aligned.columns):
            if old_aligned[col].dtype == object or new_aligned[col].dtype == object:
                object_col_indices.add(idx)

        # Fast path: no object columns means no possible NA type differences
        if not object_col_indices:
            return []

        na_type_changes: list[CellChange] = []
        na_rows, na_cols = np.where(both_na.values)

        # Use .values for fast numpy indexing instead of .iloc
        old_vals = old_aligned.values
        new_vals = new_aligned.values

        for i, j in zip(na_rows, na_cols, strict=True):
            # Skip non-object columns (they have consistent NA types)
            if j not in object_col_indices:
                continue

            old_val = old_vals[i, j]
            new_val = new_vals[i, j]
            if _get_na_type(old_val) != _get_na_type(new_val):
                # Stop collecting after MAX_CELL_CHANGES (count will still be accurate)
                if len(na_type_changes) >= MAX_CELL_CHANGES:
                    continue
                row_label = old_aligned.index[i]
                col_label = old_aligned.columns[j]
                na_type_changes.append(
                    CellChange(
                        row=row_label,
                        column=col_label,
                        old_value=old_val,
                        new_value=new_val,
                        change_type=ChangeType.NA_TYPE_CHANGED,
                    )
                )

        return na_type_changes

    def _values_equal(
        self,
        old_val: Any,
        new_val: Any,
        dtype: Any,
    ) -> bool:
        """Compare two values with appropriate tolerance.

        Args:
            old_val: Old value
            new_val: New value
            dtype: Column dtype

        Returns:
            True if values are considered equal
        """
        # Handle NA cases
        old_is_na = pd.isna(old_val)
        new_is_na = pd.isna(new_val)

        if old_is_na and new_is_na:
            return True  # Both NA, considered equal (NA type checked separately)
        if old_is_na != new_is_na:
            return False  # One NA, one not

        # Numeric comparison with tolerance
        if pd.api.types.is_numeric_dtype(dtype):
            try:
                return bool(
                    np.isclose(
                        float(old_val),
                        float(new_val),
                        rtol=self.rtol,
                        atol=self.atol,
                    )
                )
            except (ValueError, TypeError):
                return old_val == new_val

        # Exact match for other types
        return old_val == new_val

    def _detect_na_type_change(
        self,
        old_val: Any,
        new_val: Any,
    ) -> bool:
        """Detect if NA type differs between values.

        Args:
            old_val: Old value
            new_val: New value

        Returns:
            True if both are NA but of different types
        """
        old_is_na = pd.isna(old_val)
        new_is_na = pd.isna(new_val)

        if not (old_is_na and new_is_na):
            return False

        old_type = _get_na_type(old_val)
        new_type = _get_na_type(new_val)

        return old_type != new_type


def _get_na_type(val: Any) -> str:
    """Identify the type of NA value.

    Args:
        val: Value to check

    Returns:
        String identifying NA type
    """
    if val is None:
        return "None"
    if val is pd.NA:
        return "pd.NA"
    if isinstance(val, float) and np.isnan(val):
        return "np.nan"
    return "not_na"



"""Comparison module for finlab-sentinel."""

from finlab_sentinel.comparison.differ import (
    CellChange,
    ChangeType,
    ComparisonResult,
    DataFrameComparer,
)
from finlab_sentinel.comparison.hasher import ContentHasher
from finlab_sentinel.comparison.policies import (
    AppendOnlyPolicy,
    ComparisonPolicy,
    ThresholdPolicy,
)
from finlab_sentinel.comparison.report import AnomalyReport

__all__ = [
    "ContentHasher",
    "DataFrameComparer",
    "ComparisonResult",
    "CellChange",
    "ChangeType",
    "ComparisonPolicy",
    "AppendOnlyPolicy",
    "ThresholdPolicy",
    "AnomalyReport",
]



"""Data interception and orchestration."""

from __future__ import annotations

import hashlib
import logging
from collections.abc import Callable
from datetime import datetime
from typing import Any

import pandas as pd

from finlab_sentinel.comparison.differ import DataFrameComparer
from finlab_sentinel.comparison.hasher import ContentHasher
from finlab_sentinel.comparison.policies import get_policy_for_dataset
from finlab_sentinel.comparison.report import AnomalyReport
from finlab_sentinel.config.schema import SentinelConfig
from finlab_sentinel.core.hooks import get_registry as get_preprocess_registry
from finlab_sentinel.handlers.callback import create_handler_from_config
from finlab_sentinel.storage.parquet import ParquetStorage, sanitize_backup_key

logger = logging.getLogger(__name__)


class DataInterceptor:
    """Intercepts data.get calls and performs comparison logic."""

    def __init__(
        self,
        original_get: Callable,
        config: SentinelConfig,
    ) -> None:
        """Initialize interceptor.

        Args:
            original_get: Original data.get function
            config: Sentinel configuration
        """
        self.original_get = original_get
        self.config = config

        # Initialize components
        self.storage = ParquetStorage(
            base_path=config.get_storage_path(),
            compression=config.storage.compression,
        )

        self.hasher = ContentHasher()

        self.comparer = DataFrameComparer(
            rtol=config.comparison.rtol,
            atol=config.comparison.atol,
            check_dtype=config.comparison.check_dtype,
            check_na_type=config.comparison.check_na_type,
        )

        # Create handler
        self.handler = create_handler_from_config(
            behavior=config.anomaly.behavior.value,
            callback=config.anomaly.get_callback(),
        )

    def __call__(
        self,
        dataset: str,
        *args: Any,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Intercept data.get call.

        Args:
            dataset: Dataset name
            *args: Positional arguments for original function
            **kwargs: Keyword arguments for original function

        Returns:
            DataFrame (either new, cached, or as determined by handler)
        """
        logger.debug(f"Intercepting data.get for: {dataset}")

        # 1. Call original data.get
        try:
            new_data = self.original_get(dataset, *args, **kwargs)
        except Exception as e:
            logger.error(f"Original data.get failed for {dataset}: {e}")
            raise

        # Ensure it's a DataFrame
        if not isinstance(new_data, pd.DataFrame):
            # Convert FinlabDataFrame or similar to pandas DataFrame
            new_data = pd.DataFrame(new_data)

        # Keep original data to return to user
        original_data = new_data

        # 2. Apply preprocess hook if registered (for comparison only)
        preprocess_registry = get_preprocess_registry()
        new_data_for_comparison = preprocess_registry.apply(dataset, new_data)

        # 3. Generate backup key
        backup_key = self._generate_backup_key(dataset)

        # 4. Compute hash of preprocessed data
        new_hash = self.hasher.hash_dataframe(new_data_for_comparison)

        # 5. Check for existing backup (stored data is raw/original)
        cached = self.storage.load_latest(backup_key)

        if cached is None:
            # First time - save original data as baseline (not preprocessed)
            logger.info(f"First backup for {dataset}, saving as baseline")
            self.storage.save(backup_key, dataset, original_data, new_hash)
            return original_data

        cached_data, cached_metadata = cached

        # 6. Quick comparison via hash
        if new_hash == cached_metadata.content_hash:
            logger.debug(f"Hash match for {dataset}, data unchanged")
            return original_data

        logger.debug(f"Hash mismatch for {dataset}, performing full comparison")

        # 7. Full comparison - apply preprocess to cached data for comparison
        cached_for_comparison = preprocess_registry.apply(dataset, cached_data)
        result = self.comparer.compare(cached_for_comparison, new_data_for_comparison)

        if result.is_identical:
            # Hash mismatch but identical content (shouldn't happen often)
            logger.debug(f"Full comparison shows identical for {dataset}")
            return original_data

        # 8. Apply policy
        policy = get_policy_for_dataset(
            dataset=dataset,
            default_mode=self.config.comparison.policies.default_mode.value,
            history_modifiable=set(self.config.comparison.policies.history_modifiable),
            threshold=self.config.comparison.change_threshold,
            allow_na_to_value=set(self.config.comparison.policies.allow_na_to_value),
        )

        if not policy.is_violation(result):
            # Changes are within policy - update backup with original data
            logger.info(
                f"Changes accepted for {dataset}: {result.summary()} "
                f"[{policy.name} policy]"
            )
            self.storage.save(backup_key, dataset, original_data, new_hash)
            return original_data

        # 9. Policy violation - create report
        report = AnomalyReport(
            dataset=dataset,
            backup_key=backup_key,
            detected_at=datetime.now(),
            comparison_result=result,
            policy_name=policy.name,
            violation_message=policy.get_violation_message(result),
            old_hash=cached_metadata.content_hash,
            new_hash=new_hash,
        )

        # 10. Save report if configured
        if self.config.anomaly.save_reports:
            report.save(self.config.get_reports_path())

        # 11. Handle anomaly (handler may return cached or new data)
        logger.warning(f"Data anomaly detected: {report.summary}")

        # Both cached_data and original_data are raw/original data
        # Preprocess is only applied during comparison
        return self.handler.handle(report, cached_data, original_data)

    def _generate_backup_key(self, dataset: str) -> str:
        """Generate unique key for backup storage.

        Args:
            dataset: Dataset name

        Returns:
            Sanitized backup key
        """
        # Try to get universe settings from finlab
        universe_hash = self._get_universe_hash()

        return sanitize_backup_key(dataset, universe_hash)

    def _get_universe_hash(self) -> str | None:
        """Get hash of current universe settings.

        Returns:
            Hash string or None if no universe is set
        """
        try:
            # Import finlab to check universe settings
            from finlab import data

            # Try to get current universe settings
            # This depends on finlab's internal API
            if hasattr(data, "_universe") and data._universe is not None:
                universe_str = str(data._universe)
                return hashlib.md5(universe_str.encode()).hexdigest()[:8]

            # Alternative: check for universe context
            if hasattr(data, "universe") and hasattr(data.universe, "_current"):
                universe = data.universe._current
                if universe is not None:
                    universe_str = str(universe)
                    return hashlib.md5(universe_str.encode()).hexdigest()[:8]

        except Exception as e:
            logger.debug(f"Could not get universe hash: {e}")

        return None


def accept_current_data(
    dataset: str,
    config: SentinelConfig | None = None,
    reason: str | None = None,
) -> bool:
    """Accept current data as new baseline for a dataset.

    This is used to acknowledge and accept anomalous data after review.

    Args:
        dataset: Dataset name to accept
        config: Optional configuration (uses default if not provided)
        reason: Optional reason for accepting

    Returns:
        True if successful, False if dataset not found
    """
    if config is None:
        from finlab_sentinel.config.loader import load_config

        config = load_config()

    storage = ParquetStorage(
        base_path=config.get_storage_path(),
        compression=config.storage.compression,
    )

    # Generate backup key
    backup_key = sanitize_backup_key(dataset)

    # Get latest backup
    cached = storage.load_latest(backup_key)
    if cached is None:
        logger.warning(f"No backup found for {dataset}")
        return False

    # Re-fetch current data
    try:
        from finlab import data as finlab_data

        # Get the original function if we're enabled
        from finlab_sentinel.core.registry import get_original

        original_get = get_original("data.get")
        get_fn = original_get if original_get else finlab_data.get
        new_data = get_fn(dataset)

        if not isinstance(new_data, pd.DataFrame):
            new_data = pd.DataFrame(new_data)

    except Exception as e:
        logger.error(f"Failed to fetch current data for {dataset}: {e}")
        return False

    # Compute hash and save as accepted
    hasher = ContentHasher()
    new_hash = hasher.hash_dataframe(new_data)

    storage.accept_new_data(
        backup_key=backup_key,
        data=new_data,
        content_hash=new_hash,
        dataset=dataset,
        reason=reason,
    )

    logger.info(
        f"Accepted new data for {dataset}" + (f" (reason: {reason})" if reason else "")
    )

    return True



"""Tests for content hasher."""

import numpy as np
import pandas as pd

from finlab_sentinel.comparison.hasher import ContentHasher


class TestContentHasher:
    """Tests for ContentHasher class."""

    def test_identical_dataframes_same_hash(self, sample_df: pd.DataFrame):
        """Verify identical DataFrames produce same hash."""
        hasher = ContentHasher()

        hash1 = hasher.hash_dataframe(sample_df)
        hash2 = hasher.hash_dataframe(sample_df.copy())

        assert hash1 == hash2

    def test_different_values_different_hash(self, sample_df: pd.DataFrame):
        """Verify value changes produce different hash."""
        hasher = ContentHasher()

        df_modified = sample_df.copy()
        df_modified.iloc[0, 0] = 999999.0

        hash1 = hasher.hash_dataframe(sample_df)
        hash2 = hasher.hash_dataframe(df_modified)

        assert hash1 != hash2

    def test_column_order_affects_hash(self, sample_df: pd.DataFrame):
        """Verify column order affects hash."""
        hasher = ContentHasher()

        df_reordered = sample_df[sample_df.columns[::-1]]

        hash1 = hasher.hash_dataframe(sample_df)
        hash2 = hasher.hash_dataframe(df_reordered)

        assert hash1 != hash2

    def test_index_order_affects_hash(self, sample_df: pd.DataFrame):
        """Verify index order affects hash."""
        hasher = ContentHasher()

        df_reordered = sample_df.iloc[::-1]

        hash1 = hasher.hash_dataframe(sample_df)
        hash2 = hasher.hash_dataframe(df_reordered)

        assert hash1 != hash2

    def test_hash_with_na_values(self, sample_df_with_na: pd.DataFrame):
        """Verify hashing works with NA values."""
        hasher = ContentHasher()

        # Should not raise
        hash1 = hasher.hash_dataframe(sample_df_with_na)

        assert isinstance(hash1, str)
        assert len(hash1) > 0

    def test_different_na_types_different_hash(self):
        """Verify different NA types produce different hashes."""
        hasher = ContentHasher()

        # Create DataFrames with different NA types
        df_nan = pd.DataFrame({"col": [np.nan]})
        df_none = pd.DataFrame({"col": [None]})
        df_pdna = pd.DataFrame({"col": [pd.NA]})

        hash_nan = hasher.hash_dataframe(df_nan)
        hash_none = hasher.hash_dataframe(df_none)
        hash_pdna = hasher.hash_dataframe(df_pdna)

        # All should be different
        assert len({hash_nan, hash_none, hash_pdna}) == 3

    def test_hash_series(self):
        """Verify Series hashing works."""
        hasher = ContentHasher()

        series = pd.Series([1.0, 2.0, 3.0], name="test")

        hash1 = hasher.hash_series(series)
        hash2 = hasher.hash_series(series.copy())

        assert hash1 == hash2

    def test_empty_dataframe(self):
        """Verify empty DataFrame hashing works."""
        hasher = ContentHasher()

        df = pd.DataFrame()

        hash1 = hasher.hash_dataframe(df)

        assert isinstance(hash1, str)



"""Tests for preprocess hooks."""

import pandas as pd
import pytest

from finlab_sentinel.core.hooks import (
    PreprocessHookRegistry,
    clear_preprocess_hooks,
    get_registry,
    register_preprocess_hook,
    unregister_preprocess_hook,
)


@pytest.fixture
def registry() -> PreprocessHookRegistry:
    """Create a fresh registry for each test."""
    return PreprocessHookRegistry()


@pytest.fixture
def sample_df() -> pd.DataFrame:
    """Create sample DataFrame for testing."""
    return pd.DataFrame(
        {
            "A": [1.123456, 2.234567, 3.345678],
            "B": [4.456789, 5.567890, 6.678901],
        }
    )


class TestPreprocessHookRegistry:
    """Tests for PreprocessHookRegistry class."""

    def test_register_exact_hook(self, registry: PreprocessHookRegistry):
        """Verify exact match hook registration."""
        hook = lambda df: df.round(2)
        registry.register("price:收盤價", hook)

        assert registry.get("price:收盤價") is hook
        assert registry.get("price:開盤價") is None

    def test_register_pattern_hook(self, registry: PreprocessHookRegistry):
        """Verify pattern hook registration with wildcards."""
        hook = lambda df: df.round(2)
        registry.register("price:*", hook)

        assert registry.get("price:收盤價") is hook
        assert registry.get("price:開盤價") is hook
        assert registry.get("fundamental:營收") is None

    def test_exact_takes_precedence(self, registry: PreprocessHookRegistry):
        """Verify exact match takes precedence over pattern."""
        pattern_hook = lambda df: df.round(1)
        exact_hook = lambda df: df.round(2)

        registry.register("price:*", pattern_hook)
        registry.register("price:收盤價", exact_hook)

        assert registry.get("price:收盤價") is exact_hook
        assert registry.get("price:開盤價") is pattern_hook

    def test_unregister_exact(self, registry: PreprocessHookRegistry):
        """Verify unregistering exact hook."""
        hook = lambda df: df.round(2)
        registry.register("price:收盤價", hook)

        result = registry.unregister("price:收盤價")

        assert result is True
        assert registry.get("price:收盤價") is None

    def test_unregister_pattern(self, registry: PreprocessHookRegistry):
        """Verify unregistering pattern hook."""
        hook = lambda df: df.round(2)
        registry.register("price:*", hook)

        result = registry.unregister("price:*")

        assert result is True
        assert registry.get("price:收盤價") is None

    def test_unregister_nonexistent(self, registry: PreprocessHookRegistry):
        """Verify unregistering non-existent hook returns False."""
        result = registry.unregister("nonexistent")
        assert result is False

    def test_apply_with_hook(
        self, registry: PreprocessHookRegistry, sample_df: pd.DataFrame
    ):
        """Verify apply returns preprocessed DataFrame when hook exists."""
        registry.register("test", lambda df: df.round(2))

        result = registry.apply("test", sample_df)

        expected = sample_df.round(2)
        pd.testing.assert_frame_equal(result, expected)

    def test_apply_without_hook(
        self, registry: PreprocessHookRegistry, sample_df: pd.DataFrame
    ):
        """Verify apply returns original DataFrame when no hook."""
        result = registry.apply("test", sample_df)

        pd.testing.assert_frame_equal(result, sample_df)

    def test_apply_handles_error(
        self, registry: PreprocessHookRegistry, sample_df: pd.DataFrame
    ):
        """Verify apply returns original DataFrame when hook raises error."""

        def bad_hook(df: pd.DataFrame) -> pd.DataFrame:
            raise ValueError("intentional error")

        registry.register("test", bad_hook)

        result = registry.apply("test", sample_df)

        # Should return original data on error
        pd.testing.assert_frame_equal(result, sample_df)

    def test_apply_handles_wrong_return_type(
        self, registry: PreprocessHookRegistry, sample_df: pd.DataFrame
    ):
        """Verify apply returns original DataFrame when hook returns non-DataFrame."""
        registry.register("test", lambda df: "not a dataframe")

        result = registry.apply("test", sample_df)

        # Should return original data when hook returns wrong type
        pd.testing.assert_frame_equal(result, sample_df)

    def test_clear(self, registry: PreprocessHookRegistry):
        """Verify clear removes all hooks."""
        registry.register("exact", lambda df: df)
        registry.register("pattern:*", lambda df: df)

        registry.clear()

        assert registry.get("exact") is None
        assert registry.get("pattern:test") is None

    def test_list_hooks(self, registry: PreprocessHookRegistry):
        """Verify list_hooks returns all registered hooks."""
        registry.register("exact1", lambda df: df)
        registry.register("exact2", lambda df: df)
        registry.register("pattern:*", lambda df: df)

        hooks = registry.list_hooks()

        assert set(hooks["exact"]) == {"exact1", "exact2"}
        assert hooks["pattern"] == ["pattern:*"]

    def test_question_mark_wildcard(self, registry: PreprocessHookRegistry):
        """Verify ? wildcard matches single character."""
        hook = lambda df: df.round(2)
        registry.register("price:?", hook)

        assert registry.get("price:A") is hook
        assert registry.get("price:AB") is None


class TestGlobalRegistry:
    """Tests for global registry functions."""

    def setup_method(self):
        """Clear global registry before each test."""
        clear_preprocess_hooks()

    def teardown_method(self):
        """Clear global registry after each test."""
        clear_preprocess_hooks()

    def test_register_and_get(self, sample_df: pd.DataFrame):
        """Verify register and get from global registry."""
        hook = lambda df: df.round(2)
        register_preprocess_hook("test", hook)

        registry = get_registry()
        result = registry.apply("test", sample_df)

        expected = sample_df.round(2)
        pd.testing.assert_frame_equal(result, expected)

    def test_unregister(self):
        """Verify unregister from global registry."""
        register_preprocess_hook("test", lambda df: df)

        result = unregister_preprocess_hook("test")

        assert result is True
        assert get_registry().get("test") is None

    def test_clear(self):
        """Verify clear global registry."""
        register_preprocess_hook("test", lambda df: df)

        clear_preprocess_hooks()

        assert get_registry().get("test") is None


class TestPreprocessHookIntegration:
    """Integration tests for preprocess hooks with interceptor."""

    def setup_method(self):
        """Clear hooks before each test."""
        clear_preprocess_hooks()

    def teardown_method(self):
        """Clear hooks after each test."""
        clear_preprocess_hooks()

    def test_hook_applied_to_dataframe(self, sample_df: pd.DataFrame):
        """Verify hook is applied correctly."""
        # Register a hook that rounds to 2 decimal places
        register_preprocess_hook("price:*", lambda df: df.round(2))

        registry = get_registry()
        result = registry.apply("price:收盤價", sample_df)

        # Verify rounding was applied
        assert result["A"].iloc[0] == 1.12
        assert result["B"].iloc[0] == 4.46

    def test_hook_sorting_columns(self, sample_df: pd.DataFrame):
        """Verify hook can sort columns for consistent comparison."""
        df_unsorted = sample_df[["B", "A"]]

        register_preprocess_hook("test", lambda df: df[sorted(df.columns)])

        registry = get_registry()
        result = registry.apply("test", df_unsorted)

        assert list(result.columns) == ["A", "B"]



"""Pydantic configuration schema for finlab-sentinel."""

from __future__ import annotations

from collections.abc import Callable
from enum import Enum
from pathlib import Path
from typing import Literal

from pydantic import BaseModel, Field, field_validator


class PolicyMode(str, Enum):
    """Comparison policy mode."""

    APPEND_ONLY = "append_only"
    THRESHOLD = "threshold"
    PERMISSIVE = "permissive"


class AnomalyBehavior(str, Enum):
    """Behavior when anomaly is detected."""

    RAISE = "raise"
    WARN_RETURN_CACHED = "warn_return_cached"
    WARN_RETURN_NEW = "warn_return_new"


class StorageConfig(BaseModel):
    """Storage configuration."""

    path: Path = Field(default=Path("~/.finlab-sentinel/"))
    retention_days: int = Field(default=7, ge=1, le=365)
    min_backups_per_dataset: int = Field(default=3, ge=1, le=100)
    format: Literal["parquet"] = "parquet"
    compression: Literal["zstd", "snappy", "gzip", "none"] = "zstd"

    @field_validator("path", mode="before")
    @classmethod
    def expand_path(cls, v: str | Path) -> Path:
        """Expand ~ and environment variables in path."""
        if isinstance(v, str):
            v = Path(v)
        return v.expanduser()


class ComparisonPoliciesConfig(BaseModel):
    """Comparison policies configuration."""

    default_mode: PolicyMode = PolicyMode.APPEND_ONLY
    history_modifiable: list[str] = Field(default_factory=list)
    allow_na_to_value: list[str] = Field(default_factory=list)


class ComparisonConfig(BaseModel):
    """Comparison configuration."""

    rtol: float = Field(default=1e-5, ge=0)
    atol: float = Field(default=1e-8, ge=0)
    check_dtype: bool = True
    check_na_type: bool = True
    change_threshold: float = Field(default=0.10, ge=0, le=1)
    policies: ComparisonPoliciesConfig = Field(default_factory=ComparisonPoliciesConfig)


class AnomalyConfig(BaseModel):
    """Anomaly handling configuration."""

    behavior: AnomalyBehavior = AnomalyBehavior.RAISE
    save_reports: bool = True
    reports_dir: str = "reports/"
    callback: str | None = None  # "module.path:function_name"

    # Runtime callback (not serializable, set programmatically)
    _callback_fn: Callable | None = None

    model_config = {"arbitrary_types_allowed": True}

    def set_callback(self, fn: Callable) -> None:
        """Set callback function programmatically."""
        self._callback_fn = fn

    def get_callback(self) -> Callable | None:
        """Get callback function."""
        return self._callback_fn


class LoggingConfig(BaseModel):
    """Logging configuration."""

    level: Literal["DEBUG", "INFO", "WARNING", "ERROR"] = "INFO"
    file: Path | None = None

    @field_validator("file", mode="before")
    @classmethod
    def expand_file_path(cls, v: str | Path | None) -> Path | None:
        """Expand ~ in log file path."""
        if v is None:
            return None
        if isinstance(v, str):
            v = Path(v)
        return v.expanduser()


class CLIConfig(BaseModel):
    """CLI configuration."""

    output_format: Literal["table", "json", "csv"] = "table"


class SentinelConfig(BaseModel):
    """Main sentinel configuration."""

    storage: StorageConfig = Field(default_factory=StorageConfig)
    comparison: ComparisonConfig = Field(default_factory=ComparisonConfig)
    anomaly: AnomalyConfig = Field(default_factory=AnomalyConfig)
    logging: LoggingConfig = Field(default_factory=LoggingConfig)
    cli: CLIConfig = Field(default_factory=CLIConfig)

    def get_storage_path(self) -> Path:
        """Get resolved storage path."""
        return self.storage.path

    def get_reports_path(self) -> Path:
        """Get resolved reports path."""
        return self.storage.path / self.anomaly.reports_dir

    def is_dataset_history_modifiable(self, dataset: str) -> bool:
        """Check if dataset is in the history_modifiable blacklist."""
        return dataset in self.comparison.policies.history_modifiable



"""Anomaly report generation and storage."""

from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any

from finlab_sentinel.comparison.differ import (
    CellChange,
    ChangeType,
    ComparisonResult,
    DtypeChange,
)

logger = logging.getLogger(__name__)


@dataclass
class AnomalyReport:
    """Detailed report of a data anomaly."""

    dataset: str
    backup_key: str
    detected_at: datetime
    comparison_result: ComparisonResult
    policy_name: str
    violation_message: str

    # Optional metadata
    old_hash: str | None = None
    new_hash: str | None = None

    @property
    def summary(self) -> str:
        """Get short summary of the anomaly."""
        return (
            f"{self.dataset}: {self.comparison_result.summary()} "
            f"[{self.policy_name} violated]"
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert report to dictionary for serialization."""
        result = self.comparison_result

        def serialize_change(change: CellChange) -> dict:
            return {
                "row": str(change.row),
                "column": str(change.column),
                "old_value": _serialize_value(change.old_value),
                "new_value": _serialize_value(change.new_value),
                "change_type": change.change_type.value,
            }

        def serialize_dtype_change(change: DtypeChange) -> dict:
            return {
                "column": str(change.column),
                "old_dtype": change.old_dtype,
                "new_dtype": change.new_dtype,
            }

        return {
            "dataset": self.dataset,
            "backup_key": self.backup_key,
            "detected_at": self.detected_at.isoformat(),
            "policy_name": self.policy_name,
            "violation_message": self.violation_message,
            "old_hash": self.old_hash,
            "new_hash": self.new_hash,
            "comparison": {
                "is_identical": result.is_identical,
                "old_shape": list(result.old_shape),
                "new_shape": list(result.new_shape),
                "change_ratio": result.change_ratio,
                "added_rows": [str(r) for r in result.added_rows],
                "deleted_rows": [str(r) for r in result.deleted_rows],
                "added_columns": [str(c) for c in result.added_columns],
                "deleted_columns": [str(c) for c in result.deleted_columns],
                "modified_cells": [
                    serialize_change(c) for c in result.modified_cells[:100]
                ],  # Limit to 100
                "dtype_changes": [
                    serialize_dtype_change(c) for c in result.dtype_changes
                ],
                "na_type_changes": [
                    serialize_change(c) for c in result.na_type_changes[:100]
                ],
                "total_changes": result.total_changes,
            },
            "summary": self.summary,
        }

    def to_json(self, indent: int = 2) -> str:
        """Convert report to JSON string."""
        return json.dumps(self.to_dict(), indent=indent, ensure_ascii=False)

    def save(self, directory: Path) -> Path:
        """Save report to file.

        Args:
            directory: Directory to save report

        Returns:
            Path to saved report file
        """
        directory.mkdir(parents=True, exist_ok=True)

        # Generate filename
        timestamp = self.detected_at.strftime("%Y-%m-%d_%H%M%S")
        safe_key = self.backup_key.replace("/", "_").replace("\\", "_")
        filename = f"{timestamp}_{safe_key}_anomaly.json"

        filepath = directory / filename

        filepath.write_text(self.to_json(), encoding="utf-8")

        logger.info(f"Saved anomaly report to: {filepath}")

        return filepath

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> AnomalyReport:
        """Create report from dictionary.

        Args:
            data: Dictionary with report data

        Returns:
            AnomalyReport instance
        """
        comparison_data = data["comparison"]

        # Reconstruct CellChanges
        modified_cells = [
            CellChange(
                row=c["row"],
                column=c["column"],
                old_value=c["old_value"],
                new_value=c["new_value"],
                change_type=ChangeType(c["change_type"]),
            )
            for c in comparison_data.get("modified_cells", [])
        ]

        na_type_changes = [
            CellChange(
                row=c["row"],
                column=c["column"],
                old_value=c["old_value"],
                new_value=c["new_value"],
                change_type=ChangeType(c["change_type"]),
            )
            for c in comparison_data.get("na_type_changes", [])
        ]

        dtype_changes = [
            DtypeChange(
                column=c["column"],
                old_dtype=c["old_dtype"],
                new_dtype=c["new_dtype"],
            )
            for c in comparison_data.get("dtype_changes", [])
        ]

        result = ComparisonResult(
            is_identical=comparison_data["is_identical"],
            old_shape=tuple(comparison_data["old_shape"]),
            new_shape=tuple(comparison_data["new_shape"]),
            added_rows=set(comparison_data.get("added_rows", [])),
            deleted_rows=set(comparison_data.get("deleted_rows", [])),
            added_columns=set(comparison_data.get("added_columns", [])),
            deleted_columns=set(comparison_data.get("deleted_columns", [])),
            modified_cells=modified_cells,
            dtype_changes=dtype_changes,
            na_type_changes=na_type_changes,
        )

        return cls(
            dataset=data["dataset"],
            backup_key=data["backup_key"],
            detected_at=datetime.fromisoformat(data["detected_at"]),
            comparison_result=result,
            policy_name=data["policy_name"],
            violation_message=data["violation_message"],
            old_hash=data.get("old_hash"),
            new_hash=data.get("new_hash"),
        )

    @classmethod
    def load(cls, filepath: Path) -> AnomalyReport:
        """Load report from file.

        Args:
            filepath: Path to report file

        Returns:
            AnomalyReport instance
        """
        data = json.loads(filepath.read_text(encoding="utf-8"))
        return cls.from_dict(data)


def list_reports(directory: Path) -> list[Path]:
    """List all anomaly report files in directory.

    Args:
        directory: Directory to search

    Returns:
        List of report file paths, sorted by date (newest first)
    """
    if not directory.exists():
        return []

    reports = list(directory.glob("*_anomaly.json"))
    reports.sort(reverse=True)

    return reports


def _serialize_value(val: Any) -> Any:
    """Serialize a value for JSON.

    Args:
        val: Value to serialize

    Returns:
        JSON-serializable value
    """
    import numpy as np
    import pandas as pd

    # Check for numpy array first (before pd.isna which doesn't work on arrays)
    if isinstance(val, np.ndarray):
        return val.tolist()

    if isinstance(val, (np.integer, np.floating)):
        return float(val)

    if isinstance(val, (datetime, pd.Timestamp)):
        return val.isoformat()

    # Check for NA values (scalar only)
    try:
        if pd.isna(val):
            if val is None:
                return {"__na__": "None"}
            if val is pd.NA:
                return {"__na__": "pd.NA"}
            if isinstance(val, float):
                return {"__na__": "np.nan"}
            return {"__na__": "unknown"}
    except (ValueError, TypeError):
        pass  # Not a scalar NA value

    return str(val)



"""Custom exceptions for finlab-sentinel."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from finlab_sentinel.comparison.report import AnomalyReport


class SentinelError(Exception):
    """Base exception for finlab-sentinel."""


class SentinelAlreadyEnabledError(SentinelError):
    """Raised when trying to enable sentinel when it's already enabled."""


class SentinelNotEnabledError(SentinelError):
    """Raised when trying to disable sentinel when it's not enabled."""


class ConfigurationError(SentinelError):
    """Raised when there's a configuration error."""


class StorageError(SentinelError):
    """Raised when there's a storage-related error."""


class ComparisonError(SentinelError):
    """Raised when there's an error during data comparison."""


class DataAnomalyError(SentinelError):
    """Raised when a data anomaly is detected."""

    def __init__(self, report: AnomalyReport) -> None:
        self.report = report
        super().__init__(str(report))

    def __str__(self) -> str:
        return f"Data anomaly detected: {self.report.summary}"



"""Tests for monkey patching logic."""

import sys
from pathlib import Path
from types import ModuleType
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from finlab_sentinel.config.schema import SentinelConfig, StorageConfig
from finlab_sentinel.exceptions import (
    SentinelAlreadyEnabledError,
    SentinelNotEnabledError,
)


@pytest.fixture
def mock_finlab():
    """Create a mock finlab module."""
    # Create mock module
    mock_data = MagicMock()
    mock_data.get = MagicMock(return_value=pd.DataFrame({"a": [1, 2, 3]}))

    mock_finlab = ModuleType("finlab")
    mock_finlab.data = mock_data

    # Add to sys.modules
    sys.modules["finlab"] = mock_finlab
    sys.modules["finlab.data"] = mock_data

    yield mock_finlab

    # Cleanup
    if "finlab" in sys.modules:
        del sys.modules["finlab"]
    if "finlab.data" in sys.modules:
        del sys.modules["finlab.data"]


@pytest.fixture
def config_for_patcher(tmp_path: Path) -> SentinelConfig:
    """Create config for patcher tests."""
    return SentinelConfig(storage=StorageConfig(path=tmp_path))


@pytest.fixture(autouse=True)
def reset_registry():
    """Reset registry before each test."""
    from finlab_sentinel.core import registry

    registry._original_functions.clear()
    registry._enabled = False
    yield
    registry._original_functions.clear()
    registry._enabled = False


class TestEnable:
    """Tests for enable function."""

    def test_enable_patches_data_get(self, mock_finlab, config_for_patcher):
        """Verify enable patches finlab.data.get."""
        from finlab_sentinel.core.patcher import enable, is_enabled

        original_get = mock_finlab.data.get

        enable(config_for_patcher)

        # Should be enabled
        assert is_enabled()

        # data.get should be patched (not the original)
        assert mock_finlab.data.get is not original_get

    def test_enable_raises_if_already_enabled(self, mock_finlab, config_for_patcher):
        """Verify error when enabling twice."""
        from finlab_sentinel.core.patcher import enable

        enable(config_for_patcher)

        with pytest.raises(SentinelAlreadyEnabledError):
            enable(config_for_patcher)

    def test_enable_without_config_loads_default(self, mock_finlab, tmp_path):
        """Verify enable loads config from default locations."""
        from finlab_sentinel.core.patcher import enable, is_enabled

        # Create a config file in tmp_path
        # Use as_posix() to avoid Windows backslash escaping issues in TOML
        config_file = tmp_path / "sentinel.toml"
        config_file.write_text(f'[storage]\npath = "{tmp_path.as_posix()}"\n')

        # Patch the config loading to use our temp config
        with patch("finlab_sentinel.config.loader.CONFIG_SEARCH_PATHS", [config_file]):
            enable(None)

        assert is_enabled()

    def test_enable_raises_if_finlab_not_installed(self, config_for_patcher):
        """Verify error when finlab is not installed."""
        from finlab_sentinel.core.patcher import enable

        # Remove finlab from sys.modules if present
        if "finlab" in sys.modules:
            del sys.modules["finlab"]

        with pytest.raises(ImportError, match="finlab package is not installed"):
            enable(config_for_patcher)


class TestDisable:
    """Tests for disable function."""

    def test_disable_restores_original(self, mock_finlab, config_for_patcher):
        """Verify disable restores original function."""
        from finlab_sentinel.core.patcher import disable, enable, is_enabled

        original_get = mock_finlab.data.get

        enable(config_for_patcher)
        assert is_enabled()

        disable()

        assert not is_enabled()
        assert mock_finlab.data.get is original_get

    def test_disable_raises_if_not_enabled(self):
        """Verify error when disabling without enabling."""
        from finlab_sentinel.core.patcher import disable

        with pytest.raises(SentinelNotEnabledError):
            disable()


class TestIsEnabled:
    """Tests for is_enabled function."""

    def test_initially_disabled(self):
        """Verify initially returns False."""
        from finlab_sentinel.core.patcher import is_enabled

        assert not is_enabled()

    def test_returns_true_after_enable(self, mock_finlab, config_for_patcher):
        """Verify returns True after enable."""
        from finlab_sentinel.core.patcher import enable, is_enabled

        enable(config_for_patcher)
        assert is_enabled()

    def test_returns_false_after_disable(self, mock_finlab, config_for_patcher):
        """Verify returns False after disable."""
        from finlab_sentinel.core.patcher import disable, enable, is_enabled

        enable(config_for_patcher)
        disable()
        assert not is_enabled()


class TestConfigureLogging:
    """Tests for _configure_logging function."""

    def test_configures_log_level(self, config_for_patcher):
        """Verify log level is configured."""
        import logging

        from finlab_sentinel.core.patcher import _configure_logging

        config_for_patcher.logging.level = "DEBUG"
        _configure_logging(config_for_patcher)

        logger = logging.getLogger("finlab_sentinel")
        assert logger.level == logging.DEBUG

    def test_configures_file_handler(self, config_for_patcher, tmp_path):
        """Verify file handler is added when configured."""
        import logging

        from finlab_sentinel.config.schema import LoggingConfig
        from finlab_sentinel.core.patcher import _configure_logging

        log_file = tmp_path / "logs" / "sentinel.log"
        config_for_patcher.logging = LoggingConfig(file=log_file)

        _configure_logging(config_for_patcher)

        logger = logging.getLogger("finlab_sentinel")

        # Check for file handler
        file_handlers = [
            h for h in logger.handlers if isinstance(h, logging.FileHandler)
        ]
        assert len(file_handlers) >= 1

        # Cleanup handlers
        for h in logger.handlers[:]:
            logger.removeHandler(h)



"""Warning-based anomaly handlers."""

from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING

import pandas as pd

from finlab_sentinel.handlers.base import AnomalyHandler

if TYPE_CHECKING:
    from finlab_sentinel.comparison.report import AnomalyReport

logger = logging.getLogger(__name__)


class DataAnomalyWarning(UserWarning):
    """Warning issued when data anomaly is detected."""

    pass


class WarnReturnCachedHandler(AnomalyHandler):
    """Handler that warns and returns cached data."""

    @property
    def name(self) -> str:
        return "warn_return_cached"

    def handle(
        self,
        report: AnomalyReport,
        cached_data: pd.DataFrame,
        new_data: pd.DataFrame,
    ) -> pd.DataFrame:
        """Issue warning and return cached data.

        Args:
            report: Anomaly report
            cached_data: Previously cached DataFrame (returned)
            new_data: New DataFrame (discarded)

        Returns:
            The cached DataFrame
        """
        warning_msg = (
            f"Data anomaly detected for {report.dataset}: "
            f"{report.comparison_result.summary()}. "
            "Using cached data instead of new data."
        )

        warnings.warn(warning_msg, DataAnomalyWarning, stacklevel=4)
        logger.warning(warning_msg)

        return cached_data


class WarnReturnNewHandler(AnomalyHandler):
    """Handler that warns but returns new data."""

    @property
    def name(self) -> str:
        return "warn_return_new"

    def handle(
        self,
        report: AnomalyReport,
        cached_data: pd.DataFrame,
        new_data: pd.DataFrame,
    ) -> pd.DataFrame:
        """Issue warning but return new data.

        Args:
            report: Anomaly report
            cached_data: Previously cached DataFrame (discarded)
            new_data: New DataFrame (returned)

        Returns:
            The new DataFrame
        """
        warning_msg = (
            f"Data anomaly detected for {report.dataset}: "
            f"{report.comparison_result.summary()}. "
            "Returning new data despite anomaly."
        )

        warnings.warn(warning_msg, DataAnomalyWarning, stacklevel=4)
        logger.warning(warning_msg)

        return new_data



"""Anomaly handlers for finlab-sentinel."""

from finlab_sentinel.handlers.base import AnomalyHandler
from finlab_sentinel.handlers.callback import CallbackHandler
from finlab_sentinel.handlers.exception import RaiseExceptionHandler
from finlab_sentinel.handlers.warning import (
    WarnReturnCachedHandler,
    WarnReturnNewHandler,
)

__all__ = [
    "AnomalyHandler",
    "RaiseExceptionHandler",
    "WarnReturnCachedHandler",
    "WarnReturnNewHandler",
    "CallbackHandler",
]
